/* 
 * if_ppp.c - Point-to-Point Protocol (PPP) LKS for NeXT
 *
 * Copyright (c) 1989 Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by Carnegie Mellon University.  The name of the
 * University may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Drew D. Perkins
 * Carnegie Mellon University
 * 4910 Forbes Ave.
 * Pittsburgh, PA 15213
 * (412) 268-8576
 * ddp@andrew.cmu.edu
 *
 * Based on:
 *	@(#)if_sl.c	7.6.1.2 (Berkeley) 2/15/89
 *
 * Copyright (c) 1987 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of California, Berkeley.  The name of the
 * University may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Serial Line interface
 *
 * Rick Adams
 * Center for Seismic Studies
 * 1300 N 17th Street, Suite 1450
 * Arlington, Virginia 22209
 * (703)276-7900
 * rick@seismo.ARPA
 * seismo!rick
 *
 * Pounded on heavily by Chris Torek (chris@mimsy.umd.edu, umcp-cs!chris).
 * Converted to 4.3BSD Beta by Chris Torek.
 * Other changes made at Berkeley, based in part on code by Kirk Smith.
 *
 * Converted to 4.3BSD+ 386BSD by Brad Parker (brad@cayman.com)
 * Added VJ tcp header compression; more unified ioctls
 *
 * Extensively modified by Paul Mackerras (paulus@cs.anu.edu.au).
 * Cleaned up a lot of the mbuf-related code to fix bugs that
 * caused system crashes and packet corruption.  Changed pppstart
 * so that it doesn't just give up with a collision if the whole
 * packet doesn't fit in the output ring buffer.
 *
 * Added priority queueing for interactive IP packets, following
 * the model of if_sl.c, plus hooks for bpf.
 * Paul Mackerras (paulus@cs.anu.edu.au).
 *
 * Ultrix port by Per Sundstrom <sundstrom@stkhlm.enet.dec.com>,
 * Robert Olsson <robert@robur.slu.se> and Paul Mackerras.
 *
 * NeXT port by Steve Perkins <perkins@cps.msu.edu>.  Based on
 * Ultrix port by Sundstrom, Olsson, and Mackerras.
 * Converted mbuf related code to netbuf_t code.  Switched from 
 * struct ifnet to  netif_t.  Added support for LKS.
 *
 */

/* $Id: if_ppp.c,v 1.3 1995/01/14 16:51:55 perkins Exp $ */
/* from if_sl.c,v 1.11 84/10/04 12:54:47 rick Exp */

#if !defined(lint)
static char sccsid[] = "$Id: if_ppp.c,v 1.3 1995/01/14 16:51:55 perkins Exp $";
#endif /* not lint*/


#if NUM_PPP > 0

#include <kernserv/kern_server_types.h>
#import <kernserv/prototypes.h>

/* NeXT broke spl.h in 3.2/m68k. Thanks EPS! */

#if defined(m68k)
#import "spl.h"
#else
#import <kernserv/machine/spl.h>
#endif

#include "linedisc.h"   /* Not distributed in 3.2 (on Moto at least) */

#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/kernel.h>

/*
 * The import is because for some reason, HPPA
 * complains (something with multiple inclusion)
 * using normal include.
 */
#import <sys/conf.h>

#include <net/if.h>
#include <net/route.h>
#if INET
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/in_var.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#endif

#include <net/ppp_defs.h>

/*
 * HDROFF defines how many bytes to leave as a header for incoming VJ
 * compressed packets.  Basically, this allows us space to uncompress
 * the packet "in place" once we receive it.
 */
#ifdef VJC
#include <net/vjcompress.h>
#define HDROFF	(128)
#else
#define	HDROFF	(0)
#endif /* VJC */

#include <net/if_ppp.h>
#include "if_pppvar.h"
#include "ppp_async.h"


#ifdef PPP_COMPRESS
#define PACKETPTR netbuf_t
#include <net/ppp-comp.h>
#endif

/*
 * LKSs require this variable.
 */
kern_server_t instance;


#define PPP_TYPE "PPP over generic TTY" 
#define PPP_NAME "ppp"


/*
 * We're kinda based on an Ultrix kernel, we've got clists.
 * However, using double buffering on with the MUX driver
 * may defeat this.
 */

#define CCOUNT(q)	((q)->c_cc)
#define t_sc	T_LINEP   
#define	PPP_HIWAT	400	/* Don't start a new packet if HIWAT on que */


/*
 * Here is the list of our interfaces.
 */

struct ppp_softc ppp_softc[NUM_PPP];

/*
 * Prototype the functions we use for creating the 
 * virtual network interface.
 *
 * These functions are defined in this file.
 */

void pppattach __P((void));
void pppdetach __P((void));
void NB_FREE(netbuf_t nbuf);       /* eventually to become a macro */
struct ppp_softc *pppalloc __P(());
void pppdealloc __P((struct ppp_softc *sc));



/*
 * Prototype the functions we use for the Loadable
 * Line Discipline.  Not all of these functions are
 * actually called by our code.  The ttselect is the default
 * select action and is therefore not defined in 
 * any of these files.
 *
 * Most of these functions are defined in other
 * files.
 */

int     install_ld __P(());
int	pppcontrol __P((netif_t ifp, char *cmd, void *data));
int	pppioctl __P((netif_t ifp, int cmd, caddr_t data));
int	pppoutput __P((netif_t ifp, netbuf_t nbuf, void *dst));

/*
 * Places a new netbuf on sc->sc_fill_inprogress.  Not to be confused
 * with ppp_get_netbuf which returns a netbuf for other
 * applications (protocol handelers) to use.
 *
 * The top_off_recv_netbuf_queue fills up the available
 * queue of empty buffers.  Holds at most EMPTY_NETBUF_MAX
 * netbufs.  The void arg is required by kern_serv_callout().
 */

void top_off_recv_netbuf_queue(void *);
netbuf_t ppp_get_netbuf (netif_t netif);


/*
 * Now other functions needed 
 */

u_short	pppfcs __P((u_short fcs, u_char *cp, int len));
void	ppp_dump_netbuf __P((netbuf_t, int pktlen));

#ifdef PPP_COMPRESS
int ppp_ccp(struct ppp_softc *sc, netbuf_t nbuf, int rcvd);
void ppp_ccp_closed(struct ppp_softc *sc);
#endif

/*
 * A routine that compress/decompressed packets outside
 * the interrupt handler.
 */
void decompress_and_handoff_packet(void *softc);
void compress_packet(void *softc);


/*
 * The following disgusting hack gets around the problem that IP TOS
 * can't be set yet.  We want to put "interactive" traffic on a high
 * priority queue.  To decide if traffic is interactive, we check that
 * a) it is TCP and b) one of its ports is telnet, rlogin or ftp control.
 */
static u_short interactive_ports[8] = {
	0,	513,	0,	0,
	0,	21,	0,	23,
};

#define INTERACTIVE(p) (interactive_ports[(p) & 7] == (p))

/*
 * Does c need to be escaped?
 */
#define ESCAPE_P(c)	(sc->sc_asyncmap[(c) >> 5] & (1 << ((c) & 0x1F)))

#ifdef PPP_COMPRESS
/*
 * List of compressors we know about.
 * We leave some space so maybe we can modload compressors.
 *
 * List _must_ end with NULL
 *
 */

extern struct compressor ppp_bsd_compress;

struct compressor *ppp_compressors[8] = {
    &ppp_bsd_compress,
    NULL
};
#endif /* PPP_COMPRESS */



/*
 * Called when LKS is loaded to establish ppp interfaces.
 */
void pppattach()
{
    register struct ppp_softc *sc;
    register int i;
    extern int ipforwarding;
    extern int ipsendredirects;
    
    if (install_ld() < 0)
	printf("\nWarning: PPP could not install line discipline\n\n");

    printf("PPP configuring interfaces with the following defaults:\n");
    printf("  Name: %s\n", PPP_NAME);
    printf("  Type: %s\n", PPP_TYPE);
    printf("  MTU : %d\n\n", PPP_MTU);

    for (sc = ppp_softc, i=0; i < NUM_PPP; sc++, i++)
      {
	initialize_netbuf_queue(&sc->sc_in_queue ,IFQ_MAXLEN);
	initialize_netbuf_queue(&sc->sc_fast_out_queue ,IFQ_MAXLEN);
	initialize_netbuf_queue(&sc->sc_slow_out_queue ,IFQ_MAXLEN);
	initialize_netbuf_queue(&sc->sc_empty_recv_netbufs, EMPTY_NETBUF_MAX);
	initialize_netbuf_queue(&sc->sc_decomp_queue ,IFQ_MAXLEN);
	initialize_netbuf_queue(&sc->sc_compress_cache, DECOMPRESS_LOOKAHEAD);
	sc->sc_decomp_sync = 0;
	sc->sc_comp_sync = 0;

	sc->sc_send_inprogress = NULL;
	sc->sc_fill_inprogress = NULL;
	sc->sc_xfer = 0;
	sc->sc_devp = NULL;

	sc->sc_if =
	  if_attach(NULL, NULL,
		    (if_output_func_t) pppoutput,
		    (if_getbuf_func_t )ppp_get_netbuf,
		    (if_control_func_t ) pppcontrol,
		    PPP_NAME, (unsigned) i, PPP_TYPE,
		    (unsigned) PPP_MTU,
		    (unsigned) IFF_POINTOPOINT,
		    NETIFCLASS_VIRTUAL, (void *) sc);

        printf("  Unit %u... Successfully attached.\n", if_unit(sc->sc_if));

    }

    ipforwarding = 1;
    ipsendredirects = 1;

    printf("PPP-2.2 (NeXT LKS) Installed\n");
    printf("LKS Version %s\n", sccsid);
    printf("Bug reports to Steve Perkins (perkins@cps.msu.edu)\n");
}


/*
 * For NeXT, called to detach the interface
 */
void
pppdetach()
{

  netif_t ni;
  
  if (!tty_ld_remove(PPPDISC))
    printf("Warning: PPP could not remove line discipline\n");
  
  /*
   * I know we have the actual netif_ts in the sc structure,
   * but this is the suggested way and it works so I'm not
   * going to bother changing.
   */
   
  for (ni = iflist_first(); ni != NULL; ni = iflist_next(ni))
    {
      if (if_unit(ni) < NUM_PPP &&
	  (strcmp(if_type(ni),PPP_TYPE)==0) &&
	  (strcmp(if_name(ni),PPP_NAME)==0))
	{
	  printf("Detaching Interface: %s unit %d\n", if_name(ni), if_unit(ni));
	  if_flags_set(ni, IFF_POINTOPOINT);
	  if_detach(ni);

	  /*
	   * The bzero seems to be necessary
	   * to allow unloading/reloading of the
	   * LKS.  Without this, setting default
	   * routes will not work.  Have no clue
	   * as to why.  - SJP
	   */
	  bzero(ni, sizeof (netif_t));    
	}
    }
}



/*
 * Allocate a ppp interface unit and initialize it.
 */
struct ppp_softc *
pppalloc()
{
    int nppp;
    struct ppp_softc *sc;


/*
 * Transfers are currently disabled.
 */

/*
    for (nppp = 0, sc = ppp_softc; nppp < NUM_PPP; nppp++, sc++)
	if (sc->sc_xfer == pid)
	  {
	    sc->sc_xfer = 0;
	    break;
	  }


    if (nppp >= NUM_PPP)
*/


    /*
     * Search through interfaces looking for one not
     * currently being used.
     */
    
    for (nppp = 0, sc = ppp_softc; nppp < NUM_PPP; nppp++, sc++)
      {
	if (sc->sc_devp == NULL)
	    break;
      }

    /*
     * We all were used, exit.
     */

    if (nppp >= NUM_PPP)
      {
	if (sc->sc_flags & SC_DEBUG)
	  printf("Warning: no more ppp interfaces, returning NULL\n");
	return NULL;
      }

    /*
     * Else, configure the selected interface for use.
     */

    sc->sc_flags = 0;
    sc->sc_mru = PPP_MRU;
    sc->sc_fcs = 0;
    sc->sc_outfcs = 0;
    sc->sc_rawin_count = 0;

    /*
     * clear stats for pppstats
     */
    bzero(&sc->sc_stats, sizeof(struct pppstat));

#ifdef VJC
    vj_compress_init(&sc->sc_comp, -1);
#endif

#ifdef PPP_COMPRESS
    sc->sc_xc_state = NULL;
    sc->sc_rc_state = NULL;
#endif /* PPP_COMPRESS */

    if_flags_set(sc->sc_if, if_flags(sc->sc_if) | IFF_RUNNING);

    return sc;
}

/*
 * Deallocate a ppp unit.  Clear the queues and set the
 * flags to down.
 */
void pppdealloc(sc)
    struct ppp_softc *sc;
{

    free_netbuf_queue(&sc->sc_in_queue); 
    free_netbuf_queue(&sc->sc_fast_out_queue);
    free_netbuf_queue(&sc->sc_slow_out_queue);
    free_netbuf_queue(&sc->sc_empty_recv_netbufs);
    free_netbuf_queue(&sc->sc_decomp_queue);
    free_netbuf_queue(&sc->sc_compress_cache);
    sc->sc_decomp_sync = 0;
    sc->sc_comp_sync = 0;

    if (sc->sc_send_inprogress != NULL)
      {
	NB_FREE(sc->sc_send_inprogress);
	sc->sc_send_inprogress = NULL;
      }

    if (sc->sc_fill_inprogress != NULL)
      {
	NB_FREE(sc->sc_fill_inprogress);
	sc->sc_fill_inprogress = NULL;
      }

    sc->sc_xfer = 0;
    sc->sc_devp = NULL;

    if_flags_set(sc->sc_if, if_flags(sc->sc_if) & ~(IFF_UP|IFF_RUNNING));

#ifdef PPP_COMPRESS
    ppp_ccp_closed(sc); 
    sc->sc_xc_state = NULL;
    sc->sc_rc_state = NULL;
#endif /* PPP_COMPRESS */

}
    



/*
 * Queue a packet.  Start transmission if not active.
 * Packet is placed in Information field of PPP frame.
 *
 * This code is called by possibly two places.  One is from
 * the end of a TTY user level write.  The tty routines copied
 * the user data into a netbuf and as a final step, they pass
 * this netbuf to pppoutput for processing.
 *
 * The other is from a network function calling the output
 * function specified in  if_attach (i.e. pppoutput).
 *
 * In either case, the netbuf passed in has already been preshrunk
 * by PPP_HDRLEN.
 *
 */

int pppoutput(ifp, nbuf, dst)
    netif_t ifp;
    netbuf_t nbuf;
    void *dst;
{
    register struct ppp_softc *sc = &ppp_softc[if_unit(ifp)];
    struct ppp_header *ph;
    int protocol, address, control;
    u_char *cp;
    int s;  
    struct ip *ip;
    struct netbuf_queue *out_queue;

    /*
     * Try to make sure bottom half of input
     * routine has enough buffer space. It
     * cannot allocate the space itself.
     * pppoutput() is not called at interrupt
     * priority so it may call top_off_recv_netbuf_queue()
     * directly.
     */

    if (sc->sc_empty_recv_netbufs.len < EMPTY_NETBUF_MIN)
      top_off_recv_netbuf_queue((void *)sc);

    /*
     * If the interface is down and it is not a PPP protocol
     * packet, then drop packet and return error.
     */

    if (sc->sc_devp == NULL || (if_flags(ifp) & IFF_RUNNING) == 0
	|| ((if_flags(ifp) & IFF_UP) == 0
	&&((struct sockaddr *)dst)->sa_family != AF_UNSPEC ))
      {
	NB_FREE(nbuf);
	return ENETDOWN;
      }


    /*
     * Compute PPP header.
     */
    address = PPP_ALLSTATIONS;
    control = PPP_UI;

    out_queue = &sc->sc_slow_out_queue;
    
    switch (((struct sockaddr *)dst)->sa_family)
      {
#ifdef INET
    case AF_INET:
	protocol = PPP_IP;

	/*
	 * If it is an ICMP echo-request/reply then place on fast
	 * queue.
	 */

/*
	if ((ip = (struct ip *) nb_map(nbuf))->ip_p == IPPROTO_ICMP)
	    {
	      struct icmp *icmp_hdr;

	      icmp_hdr = (struct icmp *)(&(((char *)ip)[ip->ip_hl * 4]));
	      if (icmp_hdr->icmp_type == ICMP_ECHOREPLY ||
		  icmp_hdr->icmp_type == ICMP_ECHO)
		out_queue = &sc->sc_fast_out_queue;
	    }


	else
*/
	/*
	 * If this is a TCP packet to or from an "interactive" port,
	 * put the packet on the fastq instead.
	 */

	  if ((ip = (struct ip *) nb_map(nbuf))->ip_p == IPPROTO_TCP) 
	  {
	    register int p = ntohl(((int *)ip)[ip->ip_hl]);
	    if (INTERACTIVE(p & 0xffff) || INTERACTIVE(p >> 16))
		out_queue = &sc->sc_fast_out_queue;
	  }
	break;
#endif
#ifdef NS
    case AF_NS:
	protocol = PPP_XNS;
	break;
#endif
    case AF_UNSPEC:
	ph = (struct ppp_header *)((struct sockaddr *) dst)->sa_data;
	address = ph->ph_address;
	control = ph->ph_control;
	protocol = ntohs(ph->ph_protocol);
	break;
    default:
	printf("ppp%d: af%d not supported\n", if_unit(ifp), ((struct sockaddr *)dst)->sa_family);
	NB_FREE(nbuf);
	return EAFNOSUPPORT;
      }

   
    /*
     * By now, I've set my address control and protocol variables
     * and for IP traffic I have selected the queue on which
     * to place the netbuf.  All traffic goes on the slow
     * queue unless it is interactive IP traffic.
     */

    /*
     * Add PPP header.  Netbuf head was shrunk before
     * call to us.
     */

    nb_grow_top(nbuf, PPP_HDRLEN);
    
    cp = nb_map(nbuf);

    *cp++ = address;
    *cp++ = control;
    *cp++ = protocol >> 8;
    *cp++ = protocol & 0xff;

    if (sc->sc_flags & SC_LOG_OUTPKT)
      {
	printf("ppp%d output: ", if_unit(ifp));
	ppp_dump_netbuf(nbuf, -1);
      }

    /*
     * Put the packet on the appropriate queue.
     */

    if (!netbuf_enqueue(nbuf, out_queue, 1))
      {
	if (sc->sc_flags & SC_DEBUG)
	  printf("Error enquing packet at %s %d\n", __FILE__, __LINE__);
	if_oerrors_set(sc->sc_if, if_oerrors(sc->sc_if) + 1);	
	sc->sc_stats.ppp_oerrors++;
	NB_FREE(nbuf);
	return ECONNABORTED;
      }


    /*
     * If we don't have some compressed packets already
     * then do some compression.
     */
    if(sc->sc_comp_sync == 0)
    {
      s = splimp();
      if(sc->sc_comp_sync == 1)
	splx(s);
      else
	{
	  sc->sc_comp_sync = 1;
	  splx(s);
	  compress_packet((void *) sc);
	}
    }
    

    /*
     * Tell the device to send it out.
     */
/*
 * Bypass a function call here
 *
   (*sc->sc_start)(sc);
*/
    s = splimp(); 
    pppstart((struct tty *) sc->sc_devp);
    splx(s); 

    return (0);
}

/*
 * Grab another packet off a queue and apply VJ compression,
 * address/control and/or protocol compression if appropriate.
 *
 * We are called at interrupt priority.
 */

netbuf_t ppp_dequeue(sc)
    struct ppp_softc *sc;
{
  netbuf_t nbuf;
  kern_return_t success;
  int s;

  nbuf = netbuf_dequeue(&sc->sc_compress_cache, 0);

  /*
   * Schedule a callback to compress a few packets.
   * Careful with interrupt race conditions.
   */

  if(sc->sc_comp_sync == 0 &&
     (!netbuf_queue_empty_p(&sc->sc_fast_out_queue) ||
     !netbuf_queue_empty_p(&sc->sc_slow_out_queue)))
    {
/*
      s = splimp();
      if(sc->sc_comp_sync == 1)
	  splx(s);
      else
	{
*/
      if (sc->sc_flags & SC_DEBUG && nbuf == NULL)
	{
	  printf("PPP: Having to wait on callback for compression\n");
	}

      success = kern_serv_callout(&instance, compress_packet, (void*) sc);
      if (success != KERN_SUCCESS)
	{
	  if (sc->sc_flags & SC_DEBUG)
	    printf("\n\nif_ppp.c: compression callout failed...\n");
	}
      else
	sc->sc_comp_sync = 1;

/*      splx(s); */
    }
/* 		      } */
  return nbuf;
}



/*
 * PPP packet input routine.
 * The caller has checked and removed the FCS.
 * The return value is 1 if the packet was put on sc->sc_inq,
 * 0 otherwise.
 * 
 * We basically determine if it is an IP packet or some other
 * protocol (LCP, NCP...).  If IP, we hand to the kernel for processing
 * else we hand to the input queue to await a user level read().
 */

#define COMPTYPE(proto)	((proto) == PPP_VJC_COMP? TYPE_COMPRESSED_TCP: \
			 TYPE_UNCOMPRESSED_TCP)

int
ppppktin(sc, nbuf, ilen)
    struct ppp_softc *sc;
     netbuf_t nbuf;
    int ilen;
{
    kern_return_t success;

    /*
     * We do not necessarily want to call the decompression
     * routines (bsd and VJ) within an interrupt handler. So,
     * we first try to schedule a callout that will perform this
     * work in the normal thread of the LKS.  However, on those
     * (probably rare) cases where there are not enough resources,
     * we have no choice but to drop the packet on the queue and
     * have some later arriving packet schedule the callback.
     *
     * If sc->sc_decomp_sync == 1, then a callback has already
     * been scheduled.  Therfore, we just add to the queue.
     * else, we try to schedule.
     *
     * Remember, we are already at interrupt priority.
     */

    /*
     * Check compression
     */
/*    if (!sc->sc_comp_sync
	&&
	(!netbuf_queue_empty_p(&sc->sc_fast_out_queue) ||
	 !netbuf_queue_empty_p(&sc->sc_slow_out_queue)) &&
	sc->sc_compress_cache.len < DECOMPRESS_LOOKAHEAD)
*/

/*
    if (!sc->sc_decomp_sync)
      {
	success = kern_serv_callout(&instance, decompress_and_handoff_packet, (void*) sc);
	if (success != KERN_SUCCESS)
	  if (sc->sc_flags & SC_DEBUG)
	    {
	      printf("\n\nif_ppp.c: decompression callout failed...\n");
	      printf("    A later packet will schedule the callout.\n\n");
	    }
	else
	  sc->sc_decomp_sync = 1;
      }

*/
    if (!netbuf_enqueue(nbuf, &sc->sc_decomp_queue, 0))
      {
	if (sc->sc_flags & SC_DEBUG)
	  printf("Error enquing packet at %s %d\n", __FILE__, __LINE__);
	if_ierrors_set(sc->sc_if, if_ierrors(sc->sc_if) + 1);	
	sc->sc_stats.ppp_ierrors++;
	NB_FREE(nbuf);
	return 0;
      }

    if_ipackets_set(sc->sc_if, if_ipackets(sc->sc_if) + 1);
    sc->sc_stats.ppp_ipackets++;

    if (!sc->sc_decomp_sync)
      {
	success = kern_serv_callout(&instance, decompress_and_handoff_packet, (void*) sc);
	if (success != KERN_SUCCESS)
	  if (sc->sc_flags & SC_DEBUG)
	    printf("PPP: decompression callout failed. A later packet will schedule the callout.\n");
	  else
	    sc->sc_decomp_sync = 1;
      }
    return 0;
	  
}



/*
 * Process an ioctl request to interface.  There are a lot of mods
 * here.  NeXT calling structure is a tad different than 4.3.
 */


int pppcontrol(ifp, the_cmd, data1)
     register netif_t ifp;
     char *the_cmd;
     void *data1;
     
{
/*  struct ppp_softc *sc = (struct ppp_softc *) if_private(ifp); */
  if_ioctl_t *ioctl_args = NULL;
  register struct ifaddr *ifa = (struct ifaddr *)data1;
  unsigned cmd = 0;
  caddr_t data = (caddr_t)data1;
  short ifflags = *(short *)data1;
  
  if (strcmp(the_cmd, IFCONTROL_UNIXIOCTL) == 0)
    {
      ioctl_args = (if_ioctl_t *)data1;
      cmd = (unsigned) ioctl_args->ioctl_command;
      data = ioctl_args->ioctl_data;
      return pppioctl(ifp, cmd, data);
    }
  else if(strcmp(the_cmd, IFCONTROL_SETFLAGS) == 0)
    {
      if (!suser())
	return EPERM;

      if_flags_set(ifp, ifflags);

      return 0;
    }
  else if(strcmp(the_cmd, IFCONTROL_SETADDR) == 0)
    {
      struct sockaddr_in *sin = (struct sockaddr_in *)data;
      
      if (sin->sin_family != AF_INET)
	return EAFNOSUPPORT;
      return 0;
    }
  
  else if (strcmp(the_cmd, IFCONTROL_GETADDR) == 0)
    {
      if (ifa->ifa_addr.sa_family != AF_INET)
	return EAFNOSUPPORT;
    }
  else if (strcmp(the_cmd, IFCONTROL_AUTOADDR) == 0)
    {
      if (ifa->ifa_addr.sa_family != AF_INET)
	return EAFNOSUPPORT;
    }
  else
    {
      printf("Command was an unknown control string (%s)\n",the_cmd);
      return EINVAL;
    }
  
  return 0;
}


int pppioctl(ifp, cmd, data)
    register netif_t ifp;
    int cmd;
    caddr_t data;
{


    register struct ppp_softc *sc = &ppp_softc[if_unit(ifp)];
    register struct ifaddr *ifa = (struct ifaddr *)data;
    register struct ifreq *ifr = (struct ifreq *)data;
    struct ppp_stats *stats;
    struct ppp_comp_stats *cstats;
    int s = splimp(), error = 0;


    switch (cmd) {
      /*
       * Little problem with the flag length.  Should be
       * an unsigned int, but caddr_t is the best we could
       * do by using the ifru_data type.
       */

    case SIOCSIFADDR:
	if (ifa->ifa_addr.sa_family != AF_INET)
	    error = EAFNOSUPPORT;
	break;

    case SIOCSIFNETMASK:
	if (ifa->ifa_addr.sa_family != AF_INET)
	    error = EAFNOSUPPORT;
	break;

    case SIOCSIFDSTADDR:
	if (ifa->ifa_addr.sa_family != AF_INET)
	    error = EAFNOSUPPORT;
	break;

    case SIOCSIFMTU:

	if (!suser())
	  {
	    error = EPERM;
	    break;
	  }

	if(if_mtu(sc->sc_if) != ifr->ifr_mtu)
	   {
	     printf("Warning...In SIOCSIFMTU\nRequest to set MTU to %d\n",
		    (int)(ifr->ifr_mtu));
	     printf("The NeXT OS currently doesn't support changing the MTU\n");
	     printf("For now, a workaround is in place. It may break in the future.\n");
	     printf("Current value is %d\n", if_mtu(sc->sc_if));
	     printf("Attempting to change anyway...");
	     ((struct ifnet *) sc->sc_if)->if_mtu = ifr->ifr_mtu;
	     printf("  New value is %d\n", if_mtu(sc->sc_if));
	   }
	   break;

    case SIOCGIFMTU:
	ifr->ifr_mtu = if_mtu(sc->sc_if);
	break;
	
      case SIOCGPPPSTATS:
	stats = &(((struct ifpppstatsreq *) data)->stats);
	bzero(stats, sizeof(struct ppp_stats));

	stats->p.ppp_ibytes = sc->sc_stats.ppp_ibytes;
	stats->p.ppp_ipackets = sc->sc_stats.ppp_ipackets;
	stats->p.ppp_ierrors = sc->sc_stats.ppp_ierrors;
	stats->p.ppp_obytes = sc->sc_stats.ppp_obytes;
	stats->p.ppp_opackets = sc->sc_stats.ppp_opackets;
	stats->p.ppp_oerrors = 	sc->sc_stats.ppp_oerrors;
#ifdef VJC
	stats->vj.vjs_packets = sc->sc_comp.stats.vjs_packets;
	stats->vj.vjs_compressed = sc->sc_comp.stats.vjs_compressed;
	stats->vj.vjs_searches = sc->sc_comp.stats.vjs_searches;
	stats->vj.vjs_misses = sc->sc_comp.stats.vjs_misses;
	stats->vj.vjs_uncompressedin = sc->sc_comp.stats.vjs_uncompressedin;
	stats->vj.vjs_compressedin = sc->sc_comp.stats.vjs_compressedin;
	stats->vj.vjs_errorin = sc->sc_comp.stats.vjs_errorin;
	stats->vj.vjs_tossed = sc->sc_comp.stats.vjs_tossed;
#endif /* VJC */
	break;


#ifdef PPP_COMPRESS
      case SIOCGPPPCSTATS:
	cstats = &(((struct ifpppcstatsreq *) data)->stats);
	bzero(cstats, sizeof(struct ppp_comp_stats));
	if (sc->sc_xc_state != NULL)
	    (*sc->sc_xcomp->comp_stat)(sc->sc_xc_state, &cstats->c);
	if (sc->sc_rc_state != NULL)
	    (*sc->sc_rcomp->decomp_stat)(sc->sc_rc_state, &cstats->d);
	break;
#endif /* PPP_COMPRESS */
	
      default:
	printf("ioctl called with bad value of %d\n", cmd);
	error = EINVAL;
    }
    splx(s);

    return (error);
}


/*
 * When we receive compessed packets inside the interrupt
 * handler, we place the whole packet on a queue and then schedule
 * a callback to handle the decompression.  This is the callback function.
 *
 * If the packet is an IP packet we will hand it to the OS.  If it is
 * some other protocol, we will hand it to the user queue, where it will
 * eventually make it to the pppd process.
 *
 * It is possible that that new packets will be placed on the queue
 * while we are working.  So, we loop while there are packets available.
 *
 * If we place any input on the user queue, we also need to
 * wakeup those processes that are waiting.  This used to be
 * handled by the return value of ppppktin().
 *
 */

void decompress_and_handoff_packet(void *data)
{
  struct ppp_softc *sc = (struct ppp_softc *)data;
  int s, proto, rv, ilen;
  u_char *cp;
  struct ppp_header hdr;
#ifdef VJC
  u_char *iphdr;
  u_int iphlen;
  int size;
#endif
  int num_enqueued = 0;
  netbuf_t dmp = NULL, nbuf = NULL;
  
  /*
   * While I hate gotos, somethings they are useful.  It may be
   * the case that our while loop ends but the interrupt handler
   * adds a packet before we can reset the flag.  So, later
   * we may have to jump back here to handle a few extra packets.
   */
  
 decomp:
  while ((nbuf = netbuf_dequeue(&sc->sc_decomp_queue, 1)) != NULL)
    {
      cp = nb_map(nbuf);
      hdr = *(struct ppp_header *)cp;
      proto = ntohs(hdr.ph_protocol);
      
      
#ifdef PPP_COMPRESS
      /*
       * Decompress this packet if necessary, update the receiver's
       * dictionary, or take appropriate action on a CCP packet.
       */
      if (proto == PPP_COMP && sc->sc_rc_state && (sc->sc_flags & SC_DECOMP_RUN)
	  && !(sc->sc_flags & SC_DC_ERROR) && !(sc->sc_flags & SC_DC_FERROR))
	{
	  rv = (*sc->sc_rcomp->decompress)(sc->sc_rc_state, nbuf, &dmp);
	  
	  if (dmp != NULL)
	    {
	      NB_FREE(nbuf); 
	      nbuf = dmp;
	      cp = (u_char *)nb_map(nbuf);
	      proto = PPP_PROTOCOL(cp);
	    }
	  else
	    {
	      /*
	       * pass the compressed packet up to pppd, which may take
	       * CCP down or issue a Reset-Req.
	       */
	      switch (rv)
		{
		case DECOMP_OK:
		  /* no error, but no decompressed packet produced */
		  printf("Dropping decompressed packet at %s %d\n", __FILE__, __LINE__);
		  NB_FREE (nbuf); 
		  continue;
		case DECOMP_ERROR:
		  sc->sc_flags |= SC_DC_ERROR;
		  break;
		case DECOMP_FATALERROR:
		  sc->sc_flags |= SC_DC_FERROR;
		  break;
		}
	    }
	}
      else
	{
	  if (sc->sc_rc_state && (sc->sc_flags & SC_DECOMP_RUN))
	    (*sc->sc_rcomp->incomp)(sc->sc_rc_state, nbuf);
	  
	  if (proto == PPP_CCP)
	    ppp_ccp(sc, nbuf, 1);
	}
#endif
      
      
#ifdef VJC
      /*
       * See if we have a VJ-compressed packet to uncompress.
       */
      if (proto == PPP_VJC_COMP || proto == PPP_VJC_UNCOMP)
	{
	  char *pkttype = proto == PPP_VJC_COMP? "": "un";
	  
	  switch(proto)
	    {
	    case PPP_VJC_COMP:
	      if (sc->sc_flags & SC_REJ_COMP_TCP)
		{
		  if (sc->sc_flags & SC_DEBUG)
		    printf("ppp%d: %scomp pkt w/o compression; flags 0x%x\n",
			   if_unit(sc->sc_if), pkttype, sc->sc_flags);
		  if_ierrors_set(sc->sc_if, if_ierrors(sc->sc_if) + 1);
		  sc->sc_stats.ppp_ierrors++;
		  continue;
		}
	      
	      /*
	       * First, shrink this baby down to get rid
	       * of the PPP Header.  We should then be at the
	       * IP data.  We will grow it back as the last
	       * step before return.
	       */
	      nb_shrink_top(nbuf, PPP_HDRLEN);
	      size = nb_size(nbuf);
	      
	      ilen = vj_uncompress_tcp(nb_map(nbuf), size, size,
				       &sc->sc_comp, &iphdr, &iphlen); 
	      
	      if (ilen == 0)
		{
		  if_ierrors_set(sc->sc_if, if_ierrors(sc->sc_if) + 1);
		  sc->sc_stats.ppp_ierrors++;
		  
		  /*
		   * The length may be screwed up here so
		   * we reallocate to be safe.   +++ Not necessarly best way now
		   */
		  NB_FREE(nbuf);
		  continue;
		}
	      else
		{
		  /*
		   * Now set the netbuf to the appropriate place
		   *
		   * Remember, we have HDROFF (and now HDROFF + PPP_HDRLEN)
		   * extra bytes at the top of the packet.  This space
		   * was put there by pppinput() before it started
		   * to fill the packet.
		   */
		  
		  /*		ASSERT(iphlen <= HDROFF); */
		  
		  nb_grow_top(nbuf, PPP_HDRLEN + (iphlen-ilen));
		  cp = nb_map(nbuf);
		  bcopy(iphdr, cp + PPP_HDRLEN, iphlen);
		}
	      break;
	      
	    case PPP_VJC_UNCOMP:
	      nb_shrink_top(nbuf, PPP_HDRLEN);
	      if (!vj_uncompress_uncomp(nb_map(nbuf), &sc->sc_comp))
		{
		  if_ierrors_set(sc->sc_if, if_ierrors(sc->sc_if) + 1);
		  sc->sc_stats.ppp_ierrors++;
		  
		  /*
		   * The length may be screwed up here so
		   * we reallocate to be safe.   +++ Not necessarly best way now
		   */
		  NB_FREE(nbuf);
		  continue;
		}
	      nb_grow_top(nbuf, PPP_HDRLEN);
	      cp = nb_map(nbuf);
	      break;
	    }
	  
	  /*
	   * Reset the PPP framing header.
	   */
	  hdr.ph_protocol = htons(PPP_IP); 
	  *(struct ppp_header *)cp = hdr;
	  proto = PPP_IP;
	  
	}
#endif /* VJC */
      
      
      switch (proto)
	{
#ifdef INET
	case PPP_IP:
	  /*
	   * IP packet - take off the ppp header and pass it up to IP.
	   */
	  
	  if ((if_flags(sc->sc_if) & IFF_UP) == 0) 
	    {
	      /* interface is down - drop the packet. */
	      printf("Interface down... dropping packet at %s %d\n",
		     __FILE__, __LINE__);
	      NB_FREE(nbuf);
	      continue;
	    }
	  
	  
	  nb_shrink_top(nbuf, PPP_HDRLEN);
	  
	  inet_queue(sc->sc_if, nbuf); 
	  continue; 
	  break;
#endif
	  
	default:
	  /*
	   * Some other protocol - place on input queue for read().
	   */
	  break;
	}
      
      /*
       * Put the packet on the appropriate input queue.
       */
      if (!netbuf_enqueue(nbuf, &sc->sc_in_queue, 1))
	{
	  if (sc->sc_flags & SC_DEBUG)
	    printf("Error enquing packet at %s %d\n", __FILE__, __LINE__);
	  if_ierrors_set(sc->sc_if, if_ierrors(sc->sc_if) + 1);	
	  sc->sc_stats.ppp_ierrors++;
	  NB_FREE(nbuf);
	}
      else
	++num_enqueued;

    }

  /*
   * Now we have aparently emptied the queue.  So, we try to reset the
   * synchronization flag that schedules callbacks.  We check for the
   * possibility that an interrupt occurred before we finish this check.
   */
  s = splimp();
  if (!netbuf_queue_empty_p(&sc->sc_decomp_queue))
    {
      splx(s);
      goto decomp;
    }
  else
    {
      sc->sc_decomp_sync = 0;
      /*
       * Wake up those user level processes...
       * Remember we need to call ttwakeup exactly once for
       * each packet enqueued.  However, the number of packets
       * on the queue is not a valid indicator of how
       * many times to call.  So, we kept a counter...
       */

      if (num_enqueued)
	{
	  struct tty *tp = (struct tty *) sc->sc_devp;


	  if (tp == NULL || (struct ppp_softc *)tp->t_sc != sc)
	    {
	      printf("Error getting proper tp at %s %d\n", __FILE__, __LINE__);
	      if (tp == NULL)
		printf("tp was NULL\n");
	      else
		printf("tp didn't match the proper dev\n");
	    }
	  else
	    {
	      while(num_enqueued)
		{
		  putc(0, &tp->t_canq);
		  ttwakeup(tp);
		  --num_enqueued;
		}
	    }
	}
      splx(s);
    }
}


/*
 * Apply VJ compression, address/control and/or protocol compression,
 * and BSD compresion if appropriate.
 * It is the result of a callback.  We make sure that we
 * top off sc_compress_cache to a level of DECOMPRESS_LOOKAHEAD
 * before we leave.
 */

void compress_packet(void *softc)
{
  struct ppp_softc *sc = (struct ppp_softc *)softc;
  u_char *cp;
  int address, control, protocol, s;
  netbuf_t nbuf;

  while (sc->sc_compress_cache.len < DECOMPRESS_LOOKAHEAD)
    {
      nbuf = netbuf_dequeue(&sc->sc_fast_out_queue, 1);
      
      if (nbuf == NULL)
	nbuf = netbuf_dequeue(&sc->sc_slow_out_queue, 1);
      
      if (nbuf == NULL)
	{
	  sc->sc_comp_sync = 0;
	  return;
	}
      
      /*
       * Extract the ppp header of the new packet.
       */
      cp = (u_char *)nb_map(nbuf);
      
      address = cp[0];
      control = cp[1];
      protocol = (cp[2] << 8) + cp[3];
      
#ifdef VJC
      /*
       * If the packet is a TCP/IP packet, see if we can compress it.
       */
      
      if (protocol == PPP_IP && sc->sc_flags & SC_COMP_TCP)
	{
	  struct ip *ip;
	  u_int type;
	  int len;
	  u_char *vjhdr;
	  
	  ip = (struct ip *) (cp + PPP_HDRLEN);
	  
	  if (ip->ip_p == IPPROTO_TCP)
	    {
	      type = vj_compress_tcp(ip, nb_size(nbuf) - PPP_HDRLEN, &sc->sc_comp,
				     !(sc->sc_flags & SC_NO_TCP_CCID), &vjhdr);
	      
	      switch (type)
		{
		case TYPE_UNCOMPRESSED_TCP:
		  protocol = PPP_VJC_UNCOMP;
		  break;
		case TYPE_COMPRESSED_TCP:
		  protocol = PPP_VJC_COMP;
		  len = vjhdr - (u_char *) ip;
		  
		  nb_shrink_top(nbuf, len);
		  
		  cp = (u_char *) nb_map(nbuf);
		  cp[0] = address;
		  cp[1] = control;
		  cp[2] = 0;
		  break;
		case TYPE_IP:
		  break;
		}
	      if (type != TYPE_IP)
		{
		  cp[3] = protocol;	/* update protocol in PPP header */
		}
	    }
	}
#endif	/* VJC */
      
      
#ifdef PPP_COMPRESS
      if (protocol == PPP_CCP)
	ppp_ccp(sc, nbuf, 0);
#endif
      
      
#ifdef PPP_COMPRESS
      if (protocol != PPP_LCP && protocol != PPP_CCP
	  && sc->sc_xc_state && (sc->sc_flags & SC_COMP_RUN))
	{
	  netbuf_t mcomp = NULL;
	  int slen, clen;
	  
	  slen = nb_size(nbuf);
	  
	  clen = (*sc->sc_xcomp->compress)
	    (sc->sc_xc_state, &mcomp, nbuf, slen,
	     (sc->sc_flags & SC_CCP_UP? if_mtu(sc->sc_if): 0));
	  
	  
/*
	     printf("Finished compress, original size %d  comp size %d\n",
	     slen, clen);
*/
	  if (mcomp != NULL)
	    {
	      /*	    printf("Sending the compressed packet\n");  */
	      NB_FREE(nbuf);
	      nbuf = mcomp;
	      cp = (u_char *) nb_map(nbuf);
	      protocol = cp[3];
	    }
	  cp = (u_char *)nb_map(nbuf);
	}
#endif	/* PPP_COMPRESS */
      
      /*
       * Compress the address/control and protocol, if possible.
       */
      if (sc->sc_flags & SC_COMP_AC && address == PPP_ALLSTATIONS &&
	  control == PPP_UI && protocol != PPP_ALLSTATIONS &&
	  protocol != PPP_LCP)
	{
	  /* can compress address/control */
	  nb_shrink_top(nbuf, 2);
	}
      
      if (sc->sc_flags & SC_COMP_PROT && protocol < 0xFF) {
	/* can compress protocol */
	
	if ((u_char *) nb_map(nbuf) == cp) 
	  
	  {
	    cp[2] = cp[1];	/* move address/control up */
	    cp[1] = cp[0];
	  }
	nb_shrink_top(nbuf, 1);
      }
      
      s = splimp(); 

      if(!netbuf_enqueue(nbuf, &sc->sc_compress_cache, 0))
	{
	  if (sc->sc_flags & SC_DEBUG)
	    printf("Error enquing packet at %s %d\n", __FILE__, __LINE__);
	  if_oerrors_set(sc->sc_if, if_oerrors(sc->sc_if) + 1);	
	  sc->sc_stats.ppp_oerrors++;
	  NB_FREE(nbuf);
	  splx(s); 
	  break;
	}

/*
 * Bypass a function call here
 *
   (*sc->sc_start)(sc);
*/
      pppstart((struct tty *) sc->sc_devp);
      splx(s);

    }  /* While */
  sc->sc_comp_sync = 0;
}

/*
 * This returns a new netbuf for protocol handlers.
 * These buffers are typically used for sending.
 *
 */

netbuf_t ppp_get_netbuf (netif_t netif)
{
    unsigned int len = if_mtu(netif) + PPP_HDRLEN;

    netbuf_t ret_buf;

    ret_buf = nb_alloc(len);
    if (ret_buf != NULL)
	nb_shrink_top(ret_buf, (unsigned) PPP_HDRLEN);

    return ret_buf;
}



/*
 * Top off the empty netbuf queue for pppinput().
 * Must not be called from within pppinput() or
 * any other interrupt handler.  It assumes that
 * the mru has already been set.
 */

void top_off_recv_netbuf_queue(void* void_sc)
{
  struct ppp_softc *sc = (struct ppp_softc *) void_sc;
  struct netbuf_queue *queue = &sc->sc_empty_recv_netbufs;
  int len = HDROFF + sc->sc_mru + PPP_HDRLEN + PPP_FCSLEN;
  netbuf_t nbuf;



  while(queue->len < EMPTY_NETBUF_MAX)
    {
      nbuf = nb_alloc((unsigned int)len);

      if (nbuf == NULL)
	{
	  if (sc->sc_flags & SC_DEBUG)
	    printf("Trouble allocating nbufs in %s %d\n", __FILE__, __LINE__);
	  break;
	}
      else
	{
	  /*
	   * Shrink the top so we can uncompress if it is
	   * a VJC packet.
	   */
	  nb_shrink_top(nbuf, HDROFF);
	  if (!netbuf_enqueue(nbuf, queue, 1))
	    {
	      if (sc->sc_flags & SC_DEBUG)
		printf("Error enquing packet at %s %d\n", __FILE__, __LINE__);
	      NB_FREE(nbuf);
	      break;
	    }
	}
    }

}



#ifdef PPP_COMPRESS

/*
 * Handle a CCP packet.  `rcvd' is 1 if the packet was received,
 * 0 if it is about to be transmitted.
 */
int
ppp_ccp(sc, nbuf, rcvd)
    struct ppp_softc *sc;
    netbuf_t nbuf;
    int rcvd;
{
    u_char *dp, *ep;
    int slen;

    /*
     * Get a pointer to the data after the PPP header.
     */
    dp = nb_map(nbuf) + PPP_HDRLEN;
    ep = dp + (nb_size(nbuf) - PPP_HDRLEN);

    if (dp + CCP_HDRLEN > ep)
      {
	if (sc->sc_flags & SC_DEBUG)
	  printf("Warning: ppp_ccp %s packet size error at %s %d\n",
		 (rcvd == 1)?"receive":"transmit", __FILE__, __LINE__);
	return 0;
      }

    slen = CCP_LENGTH(dp);

    if (dp + slen > ep) {
	if (sc->sc_flags & SC_DEBUG)
	    printf("if_ppp/ccp: not enough data in netbuf (%x+%x > %x+%x)\n",
		   (u_int)dp, (u_int)slen, (u_int)dp, (u_int)(nb_size(nbuf) - PPP_HDRLEN));
	return 0;
    }

    switch (CCP_CODE(dp)) {
    case CCP_CONFREQ:
    case CCP_TERMREQ:
    case CCP_TERMACK:
	/* CCP must be going down - disable compression */
	if (sc->sc_flags & SC_CCP_UP)
	  {
	    printf("CCP going down, disabling compression...\n");
	    sc->sc_flags &= ~(SC_CCP_UP | SC_COMP_RUN | SC_DECOMP_RUN);
	  }
	break;

    case CCP_CONFACK:
	if (sc->sc_flags & SC_CCP_OPEN && !(sc->sc_flags & SC_CCP_UP)
	    && slen >= CCP_HDRLEN + CCP_OPT_MINLEN
	    && slen >= CCP_OPT_LENGTH(dp + CCP_HDRLEN) + CCP_HDRLEN) {
	    if (!rcvd) {
		/* we're agreeing to send compressed packets. */
		if (sc->sc_xc_state != NULL
		    && (*sc->sc_xcomp->comp_init)
			(sc->sc_xc_state, dp + CCP_HDRLEN, slen - CCP_HDRLEN,
			 if_unit(sc->sc_if), sc->sc_flags & SC_DEBUG)) {
		    sc->sc_flags |= SC_COMP_RUN;
		}
	    } else {
		/* peer is agreeing to send compressed packets. */
		if (sc->sc_rc_state != NULL
		    && (*sc->sc_rcomp->decomp_init)
			(sc->sc_rc_state, dp + CCP_HDRLEN, slen - CCP_HDRLEN,
			 if_unit(sc->sc_if), 0, sc->sc_mru,
			 sc->sc_flags & SC_DEBUG)) {
		    sc->sc_flags |= SC_DECOMP_RUN;
		    sc->sc_flags &= ~(SC_DC_ERROR | SC_DC_FERROR);
		}
	    }
	}
	break;

    case CCP_RESETACK:
	if (sc->sc_flags & SC_CCP_UP) {
	    if (!rcvd) {
		if (sc->sc_xc_state && (sc->sc_flags & SC_COMP_RUN))
		    (*sc->sc_xcomp->comp_reset)(sc->sc_xc_state);
	    } else {
		if (sc->sc_rc_state && (sc->sc_flags & SC_DECOMP_RUN)) {
		    (*sc->sc_rcomp->decomp_reset)(sc->sc_rc_state);
		    sc->sc_flags &= ~SC_DC_ERROR;
		}
	    }
	}
	break;
    }
    return 0;
}




/*
 * CCP is down; free (de)compressor state if necessary.
 */

void
ppp_ccp_closed(sc)
    struct ppp_softc *sc;
{
    if (sc->sc_xc_state) {
	(*sc->sc_xcomp->comp_free)(sc->sc_xc_state);
	sc->sc_xc_state = NULL;
    }
    if (sc->sc_rc_state) {
	(*sc->sc_rcomp->decomp_free)(sc->sc_rc_state);
	sc->sc_rc_state = NULL;
    }
}
#endif /* PPP_COMRPESS */

void NB_FREE(netbuf_t nbuf)
{
  nb_free(nbuf);
}


#endif	/* NUM_PPP > 0 */
