/*
 * if_pppvar.h - Point-to-Point Protocol variable (private) definitions.
 *
 * Copyright (c) 1989 Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by Carnegie Mellon University.  The name of the
 * University may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * $Id: if_pppvar.h,v 1.2 1995/01/14 16:51:56 perkins Exp $
 */

#ifndef _IF_PPPVAR_H_
#define _IF_PPPVAR_H_


/*
 * For this include to work properly, the KERNEL
 * flag must be defined.  However, for the
 * pppd stuff, this flag is not set and necessary
 * headers (that require the flag) are already preloaded by the time 
 * this file is loaded.  So we bypass the type if KERNEL is not set.
 *
 * This is safe because none of the pppd files reference
 * this structure slot anyway.
 *
 */

#ifndef KERNEL
typedef void* netif_t;
#endif

#include <kernserv/kern_server_types.h>
#include <net/if.h>
#include <sys/types.h>

#ifdef	VJC
#include <net/vjcompress.h>
#endif

#ifdef PPP_COMPRESS
#define PACKETPTR netbuf_t
#include <net/ppp-comp.h>
#endif

#include "nbuf_queue.h"



extern kern_server_t instance;
extern int num_out;




/*
 * Standard PPP header.
 */
struct ppp_header {
	u_char	ph_address;	/* Address Field */
	u_char	ph_control;	/* Control Field */
	u_short	ph_protocol;	/* Protocol Field */
};

/*
 * Structure describing each ppp unit.
 */
struct ppp_softc {

        netif_t sc_if;          /* The Interface structure returned by if_attach */
	u_int	sc_flags;	/* see below */
	int   sc_outstanding_pkts;  /* number of netbufs already allocated */
	void	*sc_devp;	/* pointer to device-dependent structure (our TTY struct) */
	int	(*sc_start)(struct ppp_softc *);	/* start routine */
	short	sc_mru;		/* max receive unit */
	pid_t	sc_xfer;	/* used in xferring unit to another dev */

	/* TTY side input queue, holds valid processed received 
	 * netbufs until user request with read(). ppppktin places
	 * packets here.  All packets go here except IP packets 
	 * which get processed by the kernel.
	 */
	struct netbuf_queue sc_in_queue;


	/*
	 * Since we don't want to spend time decompressing
	 * packets during the interrupt handler, we place
	 * compressed (VJ/bsd) packets here.  They
	 * are decompressed outside the interrupt handler.
	 *
	 * Communications between interrupt handler
	 * and outside thread are synchronized through
	 * sc_decomp_sync. A zero value means schedule
	 * a callout, a one means a callout has already
	 * been scheduled.
	 *
	 * Similarly for the compression except that
	 * we only compress one at a time so that
	 * we can maintain scheduling control and ordering
	 * between the prioirty queues.
	 */
	struct netbuf_queue sc_decomp_queue;
	int sc_decomp_sync;
	int sc_comp_sync;


	/*
	 * pppoutput places output packets in these queues
	 * until the asynchronous routines have a chance
	 * to send them out.  They have been encapsulated.
	 */

	struct netbuf_queue sc_fast_out_queue; /* IP interactive output packet queue */
        struct netbuf_queue sc_slow_out_queue; /* normal output packet queue */

#ifdef	VJC
	struct	vjcompress sc_comp; /* vjc control buffer */
#endif

#ifdef PPP_COMPRESS
	struct	compressor *sc_xcomp;	/* transmit compressor */
	void	*sc_xc_state;		/* transmit compressor state */
	struct	compressor *sc_rcomp;	/* receive decompressor */
	void	*sc_rc_state;		/* receive decompressor state */
#endif

	/* Device-dependent part for async lines. */
	ext_accm sc_asyncmap;	/* async control character map */
	u_int32_t sc_rasyncmap;	/* receive async control char map */

	/*
	 * Netbuf being output currently (already dequeued).
	 * pppstart uses this to store a netbuf that
	 * it couldn't finish sending to the
	 * tty.  If there is not a netbuf here (from previous
	 * call) another is dequeued from fast/slow queues. The
	 * dequeued packet may or may not end up here depending on
	 * whether we can send it all or not.
	 *
	 * sc_compress_cache is a small queue of the next few compressed buffers
         * to be sent out.  Compression is done at a lower priority.
	 * We define how many we want to store up here.
	 */
	netbuf_t sc_send_inprogress;
	struct netbuf_queue sc_compress_cache;

#define DECOMPRESS_LOOKAHEAD 5

	/*
	 * The input netbuf that is currently being received from
	 * the peer.  When it is completed, we will put it on sc_in_queue
	 * and place a new (empty) netbuf here.  Otherwise, each time
	 * a character is received (asynchronously) we place it in this
	 * netbuf.  Eventually we will get end of packet character and 
	 * then we move this.  This holder must exist because pppinput is 
	 * asynchronous and interrupt driven.  It cannot call a function to 
	 * create a new netbuf (because operation may block).
	 */

	netbuf_t sc_fill_inprogress;
	short	sc_nbuf_ilen;	       /* length of nbuf input-packet-so-far */


	/*
	 * There is a missing function in the NeXT kernel.  That
	 * is a non-blocking netbuf allocatio function.  Since
	 * we cannot allocate space inside interrupt handlers
	 * (like pppinput()) we keep a store of pre-allocated
	 * empty netbufs that these functions may use.
	 * EMPTY_NETBUF_MAX says how many to try to keep on
	 * this list.
	 *
	 * If the list is empty, characters will be dropped
	 * until some function (usually pppoutput()) fills
	 * it back up again.
	 */

#define EMPTY_NETBUF_MIN 15
#define EMPTY_NETBUF_MAX 35

	struct netbuf_queue sc_empty_recv_netbufs; 


	/*
	 * Tempraroy holders for info pertaining
	 * to current sends/receives.
	 */
	u_short	sc_fcs;		/* FCS so far (input) */
	u_short	sc_outfcs;	/* FCS so far for output packet */
	u_char	sc_rawin[16];	/* chars as received */
	int	sc_rawin_count;	/* # in sc_rawin */


	/*
	 * Support for pppstats
	 */

	struct pppstat sc_stats;

};


#endif /* _IF_PPPVAR_H_ */
