/*
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 * Netbufs don't come with nifty queuing functions like mbufs.  So make
 * some simple ones that we will use.  We keep stats on what was dropped
 * from which queue.  This is independent from the stats kept for
 * pppstats.  We may not use them, but then again... we may modify later.
 *
 * It is important to realize that allocating kernel memory is made
 * difficult by the fact that you are not allowed to call many of the
 * memory allocation routines from within interrupt handlers.  So,
 * without special handling, we would not be able to enqueue a packet
 * from with an interrupt handler.  The problem is handled by keeping a
 * list of pre-allocated buffers for use inside interrupt handlers.
 * 
 * It is unfortunate that something so trivial can get so complex.  Here,
 * I try to balance speed, resource utilization, and correct operation
 * to provide proper queueing at both normal and interrupt priority.
 *
 * Steve Perkins, 1994.
 *
 */


#include <stdio.h>

#if defined(m68k)
#import "spl.h"
#else
#import <kernserv/machine/spl.h>
#endif

#include <kernserv/kern_server_types.h>
#include <kernserv/kalloc.h>
#include "nbuf_queue.h"


extern kern_server_t instance;
void topoff_empty_netbuf_item_list(void*);    /* internal procedure */
void NB_FREE(netbuf_t nbuf);       /* eventually to become a macro */

 
/*
 * Netbufs are inserted into netbuf_items.  These netbuf_items
 * then form a linked list.  However, we must dynamically allocate
 * netbuf_items.  This, of course, causes problems if one is trying
 * to enqueue a packet at interrupt priority.  So, we keep a list of
 * empty netbuf_items around that we use.
 *
 */

static struct netbuf_item *empty_netbuf_items = NULL;
static int num_empty_netbuf_items = 0;

/*
 * We must determine when to fill up the empty_netbuf_items
 * list.  The fill procedure will probably only be called a
 * few times.  It is expected that the list will maintain an
 * average size.  Since the resources in the list are not returned
 * to the kernel, it is important to make sure to limit the number
 * we actually allocate.
 *
 */

#define ENB_HIGH_MARK 50   /* Stop here when filling */
#define ENB_LOW_MARK  20   /* Schedule refill callback if below here */ 

#define ENB_MAX_MARK  (3*ENB_HIGH_MARK)  /* Free resources after this */
					 /* mark, otherwise recycle   */


/*
 * This procedure will fill up the empty_netbuf_items list
 * to ENB_HIGH_MARK.  It _MUST NOT_ be called at interrupt
 * priority.
 *
 */

void topoff_empty_netbuf_item_list(void *arg)
{
  struct netbuf_item *head, *tail, *tmp;
  int i, s, num;

  head = (struct netbuf_item *) kalloc (sizeof(struct netbuf_item));
  if (head == NULL)
    {
      printf("Warning: memory allocation failure at %s %d\n", __FILE__, __LINE__);
      return;
    }

  tail = head;
  num = ENB_HIGH_MARK - num_empty_netbuf_items - 1;

  for (i = num; i; --i)
    {
      tmp = (struct netbuf_item *) kalloc (sizeof(struct netbuf_item));
      if (tmp == NULL)
	{
	  printf("Warning: memory allocation failure at %s %d\n", __FILE__, __LINE__);
	  break;
	}
      
      tail->next = tmp;
      tail = tmp;
    }

/*
 * Now add the linked list to the system list.
 */

  s = splimp();
  tail->next = empty_netbuf_items;
  empty_netbuf_items = head;
  num_empty_netbuf_items += num;
  splx(s);
    
}



/*
 * Initialize a netbuf queue by resetting all parameters
 * to reasonable values.  Make sure we have a reasonable 
 * number of empty_netbuf_items. If there are pre-existing
 * netbuf_items on the queue, they will become garbage.
 *
 * _MUST NOT_ be called at interrupt priority because
 * it may call topoff_empty_netbuf_item_list().
 *
 */

void initialize_netbuf_queue(struct netbuf_queue * queue, u_int max_len)
{

  int s;

  if (num_empty_netbuf_items < ENB_LOW_MARK)
    topoff_empty_netbuf_item_list(NULL);

  s = splimp();
  
  queue->max_len = max_len;
  queue->len = 0;
  queue->num_dropped = 0;
  queue->head = NULL;
  queue->tail = NULL;

  splx(s);
}
   

/*
 * Frees up and re-initializes a queue.
 *
 * Make sure you don't reset the max size.  That
 * is only set by the initialize routine. If you
 * want to reset it, call free_netbuf_queue() then
 * re-initialize with initialize_netbuf_queue().
 */

void free_netbuf_queue(struct netbuf_queue *queue)
{
  struct netbuf_item *temp;
  int s;

  s = splimp();
  
  while (queue->head != NULL)
    {
      NB_FREE(queue->head->data);
      temp = queue->head;
      queue->head = queue->head->next;

      if (num_empty_netbuf_items > ENB_MAX_MARK)
	kfree(temp, sizeof(struct netbuf_item));
      else
	{
	  temp->next = empty_netbuf_items;
	  empty_netbuf_items = temp;
	  ++num_empty_netbuf_items;
	}
    }
  queue->head = NULL;
  queue->tail = NULL;
  queue->len = 0;
  queue->num_dropped = 0;

  splx(s);
}


/*
 * If raise is true, we raise the processor
 * priority before we mess with the queue.
 * We assume that we are _not_ at interrupt
 * priority in this case.
 *
 * We return 1 if we added, else we return 0
 * if there was a problem. We leave it up to the caller
 * to detect an error return value (0) and print
 * an appropriate message/update stats.
 *
 * Note that since we may be at interrupt priority
 * we have to be careful when calling
 * topoff_empty_netbuf_item_list()
 *
 * P.S. I refuse to use gotos.  Wouldn't unwind-protect
 *      be nice.
 */

int netbuf_enqueue(netbuf_t buf, struct netbuf_queue *queue, int raise)
{
  struct netbuf_item *newitem;
  int s=0, success;

  /*
   * Make sure we have plenty of netbuf_items.
   * While there is the potential this will be
   * called many times before the callback
   * ever gets activated, emperical tests
   * show this not to be the case.
   */

  if (num_empty_netbuf_items < ENB_LOW_MARK)
    {
      if (raise)
	topoff_empty_netbuf_item_list(NULL);
      else
	{
	  success = kern_serv_callout(&instance,
				      topoff_empty_netbuf_item_list,
				      (void *) NULL);
	if (success != KERN_SUCCESS)
	    printf("Warning: Kernel server callout failed at %s %d\n",
		   __FILE__, __LINE__);
	}
    }
	


  /*
   * Get/Create wrapper and place on queue.
   */

  if (raise)
      s = splimp();


  if (queue->len >= queue->max_len)
    {
      ++queue->num_dropped;
      if (raise)
	splx(s);
      return 0;
    }



  newitem = (struct netbuf_item *) empty_netbuf_items;
  if (newitem == NULL)
    {
      printf("Warning: empty_netbuf_items list at %s %d\n",
	     __FILE__, __LINE__);
      if(raise)
	splx(s);
      return 0;
    }
  
  empty_netbuf_items = empty_netbuf_items->next;
  --num_empty_netbuf_items;
  newitem->data = buf;
  newitem->next = NULL;


  if (queue->head == NULL)
    {
      queue->head = newitem;
      queue->tail = newitem;
      queue->len = 1;
    }
  else
    {
      queue->tail->next = newitem;
      queue->tail = newitem;
      ++(queue->len);
    }

  if (raise)
    splx(s);

  return 1;
}

/*
 * If raise is true, we raise the processor
 * priority before we mess with the queue.
 */

netbuf_t netbuf_dequeue(struct netbuf_queue *queue, int raise)
{
  struct netbuf_item *item;
  netbuf_t data;
  int s=0;

  if (queue->head == NULL)
      return NULL;
  else
    {
      if(raise)
	s = splimp();

      item = queue->head;
      queue->head = queue->head->next;
      data = item->data;
      if (num_empty_netbuf_items > ENB_MAX_MARK)
	kfree(item, sizeof (struct netbuf_item));
      else
	{
	  item->next = empty_netbuf_items;
	  empty_netbuf_items = item;
	  ++num_empty_netbuf_items;
	}

      --(queue->len);


      if(raise)
	splx(s);

      return data;
    }
}


/*
 * Should probably be called at interrupt priority.
 */

int netbuf_queue_empty_p(struct netbuf_queue *queue)
{
  if (queue->head == NULL)
    {
      return 1;
    }
  else
    return 0;
}




