/* 
 * ppp_async.c - Point-to-Point Protocol (PPP) Asynchronous driver.
 *
 * Copyright (c) 1989 Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by Carnegie Mellon University.  The name of the
 * University may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Drew D. Perkins
 * Carnegie Mellon University
 * 4910 Forbes Ave.
 * Pittsburgh, PA 15213
 * (412) 268-8576
 * ddp@andrew.cmu.edu
 *
 * Based on:
 *	@(#)if_sl.c	7.6.1.2 (Berkeley) 2/15/89
 *
 * Copyright (c) 1987 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of California, Berkeley.  The name of the
 * University may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Serial Line interface
 *
 * Rick Adams
 * Center for Seismic Studies
 * 1300 N 17th Street, Suite 1450
 * Arlington, Virginia 22209
 * (703)276-7900
 * rick@seismo.ARPA
 * seismo!rick
 *
 * Pounded on heavily by Chris Torek (chris@mimsy.umd.edu, umcp-cs!chris).
 * Converted to 4.3BSD Beta by Chris Torek.
 * Other changes made at Berkeley, based in part on code by Kirk Smith.
 *
 * Converted to 4.3BSD+ 386BSD by Brad Parker (brad@cayman.com)
 * Added VJ tcp header compression; more unified ioctls
 *
 * Extensively modified by Paul Mackerras (paulus@cs.anu.edu.au).
 * Cleaned up a lot of the mbuf-related code to fix bugs that
 * caused system crashes and packet corruption.  Changed pppstart
 * so that it doesn't just give up with a collision if the whole
 * packet doesn't fit in the output ring buffer.
 *
 * Added priority queueing for interactive IP packets, following
 * the model of if_sl.c, plus hooks for bpf.
 * Paul Mackerras (paulus@cs.anu.edu.au).
 *
 * Ultrix port by Per Sundstrom <sundstrom@stkhlm.enet.dec.com>,
 * Robert Olsson <robert@robur.slu.se> and Paul Mackerras.
 *
 * NeXT port by Steve Perkins <perkins@cps.msu.edu>.  Based on
 * Ultrix port by Sundstrom, Olsson, and Mackerras.
 * Converted mbuf related code to netbuf_t code.  Switched from 
 * struct ifnet to  netif_t.  Added support for LKS.  Separated
 * asynchronous tty code from LKS code.
 *
 */

/* $Id: ppp_async.c,v 1.4 1995/01/16 19:35:51 perkins Exp $ */
/* from if_sl.c,v 1.11 84/10/04 12:54:47 rick Exp */

#if !defined(lint)
static char sccsid[] = "$Id: ppp_async.c,v 1.4 1995/01/16 19:35:51 perkins Exp $";
#endif /* not lint*/


#include <kernserv/kern_server_types.h>
#import <kernserv/prototypes.h> 
#include <sys/socket.h>
#include <sys/file.h> 


/*
 * The import is because for some reason, HPPA
 * complains (something with multiple inclusion)
 * using normal include.
 */
#import <sys/conf.h>
#include <net/if.h>
#include <net/ppp_defs.h>
#include <net/if_ppp.h> 

#include "linedisc.h"   /* Not distributed in 3.2 (on Moto at least) */

/* NeXT broke spl.h in 3.2/m68k. Thanks EPS! */

#if defined(m68k)
#import "spl.h"
#else
#import <kernserv/machine/spl.h>
#endif

#ifdef VJC
#include <netinet/in.h>
#include <netinet/in_systm.h>  
#include <netinet/ip.h>
#include <net/vjcompress.h>
#endif /* VJC */


#ifdef PPP_COMPRESS
#define PACKETPTR netbuf_t
#include <net/ppp-comp.h>

extern struct compressor ppp_bsd_compress;
extern struct compressor *ppp_compressors[8];

#endif

#include "if_pppvar.h"
#include "ppp_async.h"

#ifdef ADD_ERRORS

int error_pkt_count = 0;   /* Number of packets received */
int error_limit = -1;       /* Insert error at this limit */

#endif /* ADD_ERRORS */

/*
 * Does c need to be escaped?
 */
#define ESCAPE_P(c)	(sc->sc_asyncmap[(c) >> 5] & (1 << ((c) & 0x1F)))

#define CCOUNT(q)	((q)->c_cc)
#define	PPP_HIWAT	400	/* Don't start a new packet if HIWAT on que */
#define t_sc	T_LINEP   
#define MAX_DUMP_BYTES	128


/*extern int num_out; */
netbuf_t ppp_dequeue(struct ppp_softc*);
extern kern_server_t instance;

struct ppp_softc *pppalloc __P(());
void top_off_recv_netbuf_queue(void *);
void pppdealloc __P((struct ppp_softc *sc));
void NB_FREE(netbuf_t nbuf);       /* eventually to become a macro */
int	pppoutput __P((netif_t ifp, netbuf_t nbuf, void *dst));



/*
 * tty interface receiver interrupt.
 */
static unsigned paritytab[8] = {
    0x96696996, 0x69969669, 0x69969669, 0x96696996,
    0x69969669, 0x96696996, 0x96696996, 0x69969669
};

/*
 * FCS lookup table as calculated by genfcstab.
 */
static u_short fcstab[256] = {
	0x0000,	0x1189,	0x2312,	0x329b,	0x4624,	0x57ad,	0x6536,	0x74bf,
	0x8c48,	0x9dc1,	0xaf5a,	0xbed3,	0xca6c,	0xdbe5,	0xe97e,	0xf8f7,
	0x1081,	0x0108,	0x3393,	0x221a,	0x56a5,	0x472c,	0x75b7,	0x643e,
	0x9cc9,	0x8d40,	0xbfdb,	0xae52,	0xdaed,	0xcb64,	0xf9ff,	0xe876,
	0x2102,	0x308b,	0x0210,	0x1399,	0x6726,	0x76af,	0x4434,	0x55bd,
	0xad4a,	0xbcc3,	0x8e58,	0x9fd1,	0xeb6e,	0xfae7,	0xc87c,	0xd9f5,
	0x3183,	0x200a,	0x1291,	0x0318,	0x77a7,	0x662e,	0x54b5,	0x453c,
	0xbdcb,	0xac42,	0x9ed9,	0x8f50,	0xfbef,	0xea66,	0xd8fd,	0xc974,
	0x4204,	0x538d,	0x6116,	0x709f,	0x0420,	0x15a9,	0x2732,	0x36bb,
	0xce4c,	0xdfc5,	0xed5e,	0xfcd7,	0x8868,	0x99e1,	0xab7a,	0xbaf3,
	0x5285,	0x430c,	0x7197,	0x601e,	0x14a1,	0x0528,	0x37b3,	0x263a,
	0xdecd,	0xcf44,	0xfddf,	0xec56,	0x98e9,	0x8960,	0xbbfb,	0xaa72,
	0x6306,	0x728f,	0x4014,	0x519d,	0x2522,	0x34ab,	0x0630,	0x17b9,
	0xef4e,	0xfec7,	0xcc5c,	0xddd5,	0xa96a,	0xb8e3,	0x8a78,	0x9bf1,
	0x7387,	0x620e,	0x5095,	0x411c,	0x35a3,	0x242a,	0x16b1,	0x0738,
	0xffcf,	0xee46,	0xdcdd,	0xcd54,	0xb9eb,	0xa862,	0x9af9,	0x8b70,
	0x8408,	0x9581,	0xa71a,	0xb693,	0xc22c,	0xd3a5,	0xe13e,	0xf0b7,
	0x0840,	0x19c9,	0x2b52,	0x3adb,	0x4e64,	0x5fed,	0x6d76,	0x7cff,
	0x9489,	0x8500,	0xb79b,	0xa612,	0xd2ad,	0xc324,	0xf1bf,	0xe036,
	0x18c1,	0x0948,	0x3bd3,	0x2a5a,	0x5ee5,	0x4f6c,	0x7df7,	0x6c7e,
	0xa50a,	0xb483,	0x8618,	0x9791,	0xe32e,	0xf2a7,	0xc03c,	0xd1b5,
	0x2942,	0x38cb,	0x0a50,	0x1bd9,	0x6f66,	0x7eef,	0x4c74,	0x5dfd,
	0xb58b,	0xa402,	0x9699,	0x8710,	0xf3af,	0xe226,	0xd0bd,	0xc134,
	0x39c3,	0x284a,	0x1ad1,	0x0b58,	0x7fe7,	0x6e6e,	0x5cf5,	0x4d7c,
	0xc60c,	0xd785,	0xe51e,	0xf497,	0x8028,	0x91a1,	0xa33a,	0xb2b3,
	0x4a44,	0x5bcd,	0x6956,	0x78df,	0x0c60,	0x1de9,	0x2f72,	0x3efb,
	0xd68d,	0xc704,	0xf59f,	0xe416,	0x90a9,	0x8120,	0xb3bb,	0xa232,
	0x5ac5,	0x4b4c,	0x79d7,	0x685e,	0x1ce1,	0x0d68,	0x3ff3,	0x2e7a,
	0xe70e,	0xf687,	0xc41c,	0xd595,	0xa12a,	0xb0a3,	0x8238,	0x93b1,
	0x6b46,	0x7acf,	0x4854,	0x59dd,	0x2d62,	0x3ceb,	0x0e70,	0x1ff9,
	0xf78f,	0xe606,	0xd49d,	0xc514,	0xb1ab,	0xa022,	0x92b9,	0x8330,
	0x7bc7,	0x6a4e,	0x58d5,	0x495c,	0x3de3,	0x2c6a,	0x1ef1,	0x0f78
};




/*
 * A little undocumented magic here.  Got the header
 * file "linedisc.h" from NeXT.  There is no documentation
 * on the functions, but they are basically similar to 
 * 4.3 line disciplines.
 */


int install_ld()
{

    return tty_ld_install(PPPDISC, NORMAL_LDISC, pppopen,
			  pppclose, pppread, pppwrite, ppptioctl,
			  pppinput, ppprend, pppstart, ttymodem,
			  ttselect);
}



/*
 * Line specific open routine for async tty devices.
 * Attach the given tty to the first available ppp unit.
 *
 * This function is invoked when a user level routine calles
 * open() and the line dicipline has been set to PPPDISC.
 * We first allocate a softc structure for this user.
 *
 * We assume that everything should be initialized
 * here.  Either this is the first open call which
 * means no previous initilization has been done, or
 * it is a not first in which case, pppclose() should
 * have cleaned up for us.
 */

int pppopen(dev, tp)
    dev_t dev;
    register struct tty *tp;
{
    register struct ppp_softc *sc;

    if (!suser())
	return EPERM;

    /* Check to see if already open */
    if (tp->t_line == PPPDISC)
      {
	sc = (struct ppp_softc *) tp->t_sc;
	if (sc != NULL && sc->sc_devp == (void *) tp)
	    return (0);
      }

    /*
     * pppalloc returns an initialied ppp_softc
     * structure for use.
     */
    if ((sc = pppalloc()) == NULL)
	return ENXIO;


    /*
     * Make sure pppinput() has some buffers to use.
     */

    top_off_recv_netbuf_queue((void *)sc);


    if ((sc->sc_fill_inprogress = netbuf_dequeue(&sc->sc_empty_recv_netbufs, 1)) == NULL)
      {
	if (sc->sc_flags & SC_DEBUG)
	  printf("Out of buffers at at line %d of file %s\n",
		 __LINE__,__FILE__);

	if_flags_set(sc->sc_if, if_flags(sc->sc_if) & ~(IFF_UP|IFF_RUNNING));

	return (ENOBUFS);
      }


    sc->sc_nbuf_ilen = 0;

    bzero(sc->sc_asyncmap, sizeof(sc->sc_asyncmap));
    sc->sc_asyncmap[0] = 0xffffffff;
    sc->sc_asyncmap[3] = 0x60000000;
    sc->sc_rasyncmap = 0;
    sc->sc_devp = (void *) tp;
    sc->sc_start = pppasyncstart;
    tp->t_sc = (caddr_t) sc;

    ttyflush(tp, FREAD | FWRITE);

    return (0);
}

/*
 * Line specific close routine.
 * Detach the tty from the ppp unit.
 * Mimics part of ttyclose().
 *
 * Invoked on user level close().
 *
 */
void
pppclose(tp)
    struct tty *tp;
{
  register struct ppp_softc *sc;
  int s;
  
  ttywflush(tp);
  s = splimp();		/* paranoid; splnet probably ok */
  tp->t_line = 0;       /* i.e. not PPPDISC             */
  sc = (struct ppp_softc *)tp->t_sc;
  if (sc != NULL)
    {
      tp->t_sc = NULL;
      if (tp == (struct tty *) sc->sc_devp)
	pppdealloc(sc);
      else
	printf("Warning, not deallocating netif %s %d\n",
	       __FILE__, __LINE__);
    }
  else
    printf("Warning, close called on non-PPP %s %d\n",
	   __FILE__, __LINE__);
  splx(s);
}



/*
 * Line specific (tty) read routine.
 *
 * Invoked on user level read(). Move data recevied from
 * modem and waiting on TTY input queue to user space.
 *
 */

int pppread(tp, uio)
    register struct tty *tp;
    struct uio *uio;
{
    register struct ppp_softc *sc = (struct ppp_softc *)tp->t_sc;
    register int s;
    int error = 0, nbuf_len;
    netbuf_t nbuf;


    if ((tp->t_state & TS_CARR_ON) == 0 && (tp->t_flags & CLOCAL) == 0)
	return 0;		/* end of file */

    if (sc == NULL || tp != (struct tty *) sc->sc_devp)
	return 0;

    s = splimp(); 

    while (netbuf_queue_empty_p(&sc->sc_in_queue) && tp->t_line == PPPDISC) 
      {
	if (tp->t_state & (TS_ASYNC | TS_NBIO))
	  {
	    splx(s); 
	    return (EWOULDBLOCK);    /* Dont sleep if process requested Asynchronous */
	  }
	sleep((caddr_t) &tp->t_rawq, TTIPRI);  /* Sleep until more arrival */
      }

    if (tp->t_line != PPPDISC)
      {
	splx(s); 
	return (-1);
      }

    
    /*
     * Pull place-holder byte out of canonical queue that was
     * put there by pppinput() (or decompress_and_handoff_packet())
     * just before it woke you up
     * (you, meaning pppread, are an upper half routine).
     */
    getc(&tp->t_canq);

    /* Get the packet from the input queue */

    nbuf = netbuf_dequeue(&sc->sc_in_queue, 1);
    splx(s);

    if (nbuf == NULL)
      {
	printf("Read didn't get a buffer at %s %d\n", __FILE__, __LINE__);
	return -1;
      }

    nbuf_len = nb_size(nbuf);
    if (error = uiomove(nb_map(nbuf), nbuf_len, UIO_READ, uio))
      printf("Error in uiomove\n");

    NB_FREE(nbuf);

    return (error);
}


/*
 * Line specific (tty) write routine.
 *
 *  Invoked on user level write().
 *
 *  Here we take user level output and copy to kernel space.
 *  Then we place on output queue and start up a sending process.
 *
 */

int pppwrite(tp, uio)
    register struct tty *tp;
    struct uio *uio;
{
    register struct ppp_softc *sc = (struct ppp_softc *)tp->t_sc;
    struct sockaddr dst;
    struct ppp_header *ph1, *ph2;
    int  error;
    netbuf_t nbuf;
    

    if ((tp->t_state & TS_CARR_ON) == 0 && (tp->t_flags & CLOCAL) == 0)
	return 0;		/* wrote 0 bytes */

    if (tp->t_line != PPPDISC)
	return (EINVAL);

    if (sc == NULL || tp != (struct tty *) sc->sc_devp)
	return EIO;

    if (uio->uio_resid > if_mtu(sc->sc_if) + PPP_HDRLEN ||
	uio->uio_resid < PPP_HDRLEN)
	return (EMSGSIZE);


    /*
     * Always try to allocate netbufs for PPP protcol
     * packets.  We don't want this protocol to be
     * goverened by the higher level packet limit.
     */
    
    nbuf = nb_alloc((unsigned int) uio->uio_resid);

    if (nbuf == NULL)
      {
	if (sc->sc_flags & SC_DEBUG)
	  printf("No buffers available for user level write()\n");
	return(ENOBUFS);
      }

    if (error = uiomove(nb_map(nbuf), uio->uio_resid, UIO_WRITE, uio))
      {
	NB_FREE(nbuf);
	return (error);
      }
    
    dst.sa_family = AF_UNSPEC;
    ph1 = (struct ppp_header *) dst.sa_data;
    ph2 = (struct ppp_header *) nb_map(nbuf);
    *ph1 = *ph2;

    nb_shrink_top(nbuf, PPP_HDRLEN);

    return (pppoutput(sc->sc_if, nbuf, &dst));
}


/*
 * Line specific (tty) ioctl routine.
 * Provide a way to get the ppp unit number.
 * This discipline requires that tty device drivers call
 * the line specific l_ioctl routine from their ioctl routines.
 *
 * Basically allows querying/setting the fields a ppp_softc;
 */


int ppptioctl(tp, cmd, data, flag)
    struct tty *tp;
    int cmd;
    void *data;
    int flag;
{
  register struct ppp_softc *sc = (struct ppp_softc *) tp->t_sc;
  int s, error, flags, mru;
/*  u_int tmp_val; */

#ifdef	i386
    struct uthread *_u = (struct uthread*) uthread_from_thread(current_thread());
#endif

#ifdef PPP_COMPRESS
    int nb;
    struct ppp_option_data *odp;
    struct compressor **cp;
    u_char ccp_option[CCP_MAX_OPTION_LENGTH];
#endif

  if (sc == NULL || tp != (struct tty *) sc->sc_devp)
    {
      printf("PPP Warning: Bad value at %s %d\n", __FILE__, __LINE__);
      return -1;
    }

  error = 0;

  switch (cmd)
    {

      /*
       * There are bugs in the way ioctls are handled on 
       * intel systems.  Unfortunatly, an ioctl that is defined
       * as _IOR or _IOWR will _not_ have its value copied back into
       * user space.  The solution is to pass in the virtual address of
       * the space you want to write to.  Then, we manually map that to
       * kernel space, perform the operation, then manually copy it back.
       *
       * This bug only seems to appear on the line disciplines.
       * Socket ioctls seem to work fine.
       *
       * Note that since things do work correctly on Black, and since
       * the ioctl must be redefined as _IOWR for Intel, we must both
       * copy back the value manually (will work for Intel) and save the
       * value in 'data' since black will copy this value back.  This covers
       * the case where the caller has placed his own address as his value.
       * Further, if the system does start to work properly in the future, 
       * no changes will be _required_ here.  Of course, cleanup would be nice.
       *
       */


    case PPPIOCGUNIT:
/*
      tmp_val = if_unit(sc->sc_if);
      copyout(&tmp_val, (void *) *(u_int *)data, sizeof(int));
*/
      *(u_int *)data = if_unit(sc->sc_if);
      break;

    case PPPIOCGFLAGS:
/*
      tmp_val = sc->sc_flags;
      copyout(&tmp_val, (void *) *(u_int *)data, sizeof(int));
*/
      *(u_int *)data = sc->sc_flags;
      break;



    case PPPIOCSFLAGS:
      if (!suser())
	return EPERM;

      flags = *(u_int *)data & SC_MASK;
      s = splimp();
      if (sc->sc_flags & SC_CCP_OPEN && !(flags & SC_CCP_OPEN))
	ppp_ccp_closed(sc);
      sc->sc_flags = (sc->sc_flags & ~SC_MASK) | flags;
      splx(s);
      break;

    case PPPIOCSASYNCMAP:
      if (!suser())
	return EPERM;
      sc->sc_asyncmap[0] = *(u_int *)data;
      break;

/*
    case PPPIOCGASYNCMAP:
      tmp_val = sc->sc_asyncmap[0];
      copyout(&tmp_val, (void *) *(u_int *)data, sizeof(int));
      break;
*/

    case PPPIOCSRASYNCMAP:
      if (!suser())
	return EPERM;
      sc->sc_rasyncmap = *(u_int *)data;
      break;

/*
    case PPPIOCGRASYNCMAP:
      tmp_val = sc->sc_rasyncmap;
      copyout(&tmp_val, (void *) *(u_int *)data, sizeof(int));
      break;
*/

    case PPPIOCSXASYNCMAP:
      if (!suser())
	return EPERM;
      bcopy(data, sc->sc_asyncmap, sizeof(sc->sc_asyncmap));
      sc->sc_asyncmap[1] = 0;		    /* mustn't escape 0x20 - 0x3f */
      sc->sc_asyncmap[2] &= ~0x40000000;  /* mustn't escape 0x5e */
      sc->sc_asyncmap[3] |= 0x60000000;   /* must escape 0x7d, 0x7e */
      break;

/*
    case PPPIOCGXASYNCMAP:
      copyout(&sc->sc_asyncmap, (void *) *(u_int *)data, sizeof(sc->sc_asyncmap));
      break;
*/

    case PPPIOCSMRU:
      if (!suser())
	return EPERM;
      mru = *(int *)data;
      if (mru >= PPP_MRU && mru <= PPP_MAXMRU)
	{
	  s = splimp();

	  sc->sc_mru = mru;

	  /*
	   * Since we changed MRU, make sure our
	   * netbufcan handle data of the new
	   * (possibly larger) length.
	   */
	  
	  free_netbuf_queue(&sc->sc_empty_recv_netbufs);
	  top_off_recv_netbuf_queue((void *)sc);
	  
	  if (sc->sc_fill_inprogress)
	    NB_FREE(sc->sc_fill_inprogress);

	  if ((sc->sc_fill_inprogress = netbuf_dequeue(&sc->sc_empty_recv_netbufs, 1)) == NULL)
	    {
	      if (sc->sc_flags & SC_DEBUG)
		printf("Out of buffers at line %d of file %s\n",
		       __LINE__,__FILE__);
	      sc->sc_mru = PPP_MRU;
	      
	      /*
	       * On failure, try for the default PPP_MRU size.
	       */
	      
	      free_netbuf_queue(&sc->sc_empty_recv_netbufs);
	      top_off_recv_netbuf_queue((void *)sc);
	      
	      if ((sc->sc_fill_inprogress = netbuf_dequeue(&sc->sc_empty_recv_netbufs, 1)) == NULL)
		{
		  if (sc->sc_flags & SC_DEBUG)
		    printf("Out of buffers at line %d of file %s\n",
			   __LINE__,__FILE__);

		  if_flags_set(sc->sc_if, if_flags(sc->sc_if) & ~IFF_UP);
		  sc->sc_flags |= SC_FLUSH;
		  error = ENOBUFS;
		}
	    }
	  sc->sc_nbuf_ilen = 0;
	  splx(s);
	}
      break;

/*
    case PPPIOCGMRU:
      tmp_val = sc->sc_mru;
      copyout(&tmp_val, (void *) *(u_int *)data, sizeof(int));
      break;
*/

#ifdef VJC
    case PPPIOCSMAXCID:
      if (!suser())
	return EPERM;
      vj_compress_init(&sc->sc_comp, *(int *)data);
      break;
#endif

/*
 * Disable transferring units for now
 */
/*      
    case PPPIOCXFERUNIT:
      if (!suser())
	return EPERM;
      sc->sc_xfer = p->p_pid;
      break;
*/

      /*
       * Set up support for CCP packet data compression.
       */

#ifdef PPP_COMPRESS
    case PPPIOCSCOMPRESS:
      if (!suser())
	return EPERM;

	odp = (struct ppp_option_data *) data;
	nb = odp->length;
	if (nb > sizeof(ccp_option))
	    nb = sizeof(ccp_option);
	if (error = copyin(odp->ptr, ccp_option, nb))
	    return (error);
	if (ccp_option[1] < 2)	/* preliminary check on the length byte */
	    return (EINVAL);
	for (cp = ppp_compressors; *cp != NULL; ++cp)
	    if ((*cp)->compress_proto == ccp_option[0]) {
		/*
		 * Found a handler for the protocol - try to allocate
		 * a compressor or decompressor.
		 */

	      if (sc->sc_flags & SC_DEBUG)
		{
		  if(odp->transmit)
		    printf("PPP Allocating BSD Transmit compressor...\n");
		  else
		    printf("PPP Allocating BSD Receive decompressor...\n");
		}

		error = 0;
		s = splimp();
		if (odp->transmit) {
		    if (sc->sc_xc_state != NULL)
			(*sc->sc_xcomp->comp_free)(sc->sc_xc_state);
		    sc->sc_xcomp = *cp;
		    sc->sc_xc_state = (*cp)->comp_alloc(ccp_option, nb);
		    if (sc->sc_xc_state == NULL) {
			if (sc->sc_flags & SC_DEBUG)
			    printf("ppp%d: comp_alloc failed\n",
				   if_unit(sc->sc_if));
			error = ENOBUFS;
		    }
		    sc->sc_flags &= ~SC_COMP_RUN;
		} else {
		    if (sc->sc_rc_state != NULL)
			(*sc->sc_rcomp->decomp_free)(sc->sc_rc_state);
		    sc->sc_rcomp = *cp;
		    sc->sc_rc_state = (*cp)->decomp_alloc(ccp_option, nb);
		    if (sc->sc_rc_state == NULL) {
			if (sc->sc_flags & SC_DEBUG)
			    printf("ppp%d: decomp_alloc failed\n",
				   if_unit(sc->sc_if));
			error = ENOBUFS;
		    }
		    sc->sc_flags &= ~SC_DECOMP_RUN;
		}
		splx(s);
		return (error);
	    }
	if (sc->sc_flags & SC_DEBUG)
	    printf("ppp%d: no compressor for [%x %x %x], %x\n",
		   if_unit(sc->sc_if), ccp_option[0], ccp_option[1],
		   ccp_option[2], nb);
	return (EINVAL);	/* no handler found */
      break;
#endif /* PPP_COMPRESS */
      
    default:
      return (-1);
    }

/*
 * Thanks to Philippe-Andre Prindeville <philipp@res.enst.fr>
 */
#ifdef	i386
    if (! error && (cmd & IOC_OUT))
	copyout(data, (caddr_t) _u->uu_arg[2], (cmd >> 16) & IOCPARM_MASK);
#endif


  return (0);
}



/*
 * 4.3 says this is an unused function.  However,
 * it appears to be returning a NULL terminated string
 * of several characters.  My guess is that the serial
 * driver is doing a little buffering so that we don't
 * get burdend with interrupts.  Just pass them off.
 * Great place for an optimization. 
 *
 * This function gets called when you use the NeXT
 * supplied serial drivers.  It does not get called
 * with the MuX driver.  
 *
 * In order to expedite the work done here, we
 * handle most things here that don't require
 * processing of a PPP_FLAG.
 *
 * The outline for this section of code (really an optimization
 * from the old function) was taken from Miron Cuperman's ppp-0.3
 * function.  Thanks Miron!
 */

void ppprend(char *cp, u_int n, struct tty *tp)
{
  register struct ppp_softc *sc = (struct ppp_softc *)tp->t_sc;
  register int ret;

  if (sc == NULL || tp != (struct tty *) sc->sc_devp)
    {
      printf("Warning, bad softc structure at %s %d\n", __FILE__, __LINE__);
      return;
    }

  /*
   * We can handle FLUSHs, ESCAPES, and non PPP_FLAG characters
   */
  
  while (n)
    {
      if (sc->sc_flags & SC_FLUSH)
	{
	  do
	    {
	      if (*(cp++) == PPP_FLAG)
		{
		  pppinput(PPP_FLAG, tp);
		  --n;
		  break;
		}
	      else if (sc->sc_flags & SC_LOG_FLUSH)
		ppplogchar(sc, *cp);
	    }
	  while(--n);
	}
      else if (sc->sc_nbuf_ilen > 3 &&
		       nb_size(sc->sc_fill_inprogress) > n &&
		 *cp != PPP_FLAG &&
		 *cp != PPP_ESCAPE)
	{
	    char* cp1 = cp;
	    if (sc->sc_flags & SC_ESCAPED)
	    {
		sc->sc_flags &= ~SC_ESCAPED;
		*cp ^= PPP_TRANS;
	    }
	    
	    do
	    {
		sc->sc_fcs = PPP_FCS(sc->sc_fcs, *(cp++));
	    } while(--n && *cp != PPP_FLAG && *cp != PPP_ESCAPE);
	    

	    if ((ret = nb_write(sc->sc_fill_inprogress, 0, (cp-cp1), cp1)))
	      printf("nb_write failed: ret = %x\n", ret);

	    sc->sc_stats.ppp_ibytes += (cp - cp1);
	    sc->sc_nbuf_ilen += (cp-cp1);
	    nb_shrink_top(sc->sc_fill_inprogress, (cp-cp1));
	}
	else
	{
	    --n;
	    pppinput(*(cp++), tp);
	}
    }
}


/*
 *  Called when a character recevied from Modem
 *  It is a bottom half routine.
 *
 *  It adds the character to the netbuf in sc_fill_inprogress
 *  and if the character marks the end of a valid
 *  packet, it wakes up those people that may be sleeping
 *  waiting on input.
 *
 * There is one problem with netbuf usage.  You cannot allocate
 * netbufs inside interrupt handlers because nb_alloc() may block.
 * So, if we fill up a buffer, what do we do?  There is no way to
 * allocate one inside pppinput. This would be OK as long as we could
 * guarantee that if we use up a netbuf, somebody else will replace it
 * before the next interrupt.  However, as is obvious, this routine is
 * asynchronous and there is no function that is guaranteed to be called
 * between invocations of this function.  We do our best... we keep a
 * queue of pre-allocated empty netbufs waiting.  If the queue empties
 * out before someone refills it, we must drop characters/packets until
 * somebody gets around to refilling the queue.
 *
 * It is important to understand how the empty buffers work.
 * The first time this function is entered, we are guaranteed
 * to have a good (non-NULL non-full) netbuf waiting on
 * sc->sc_fill_inprogress.  This is guaranteed through pppopen().
 * We also have empty netbufs on the sc->sc_empty_recv_netbufs
 * queue.  If we ever run out of netbufs inside this routine, we
 * first try to allocate from the empty netbuf queue.  If that fails,
 * we flush characters until we can get a new empty netbuf (i.e. somebody
 * refilled the queue.  If we ever set sc_fill_inprogress to NULL then
 * we must also set the SC_FLUSH flag.  
 * 
*/

void pppinput(c, tp)
     int c;
     register struct tty *tp;
{
  register struct ppp_softc *sc;
  netbuf_t nbuf; 
  u_char *cur_pos;
  int nbuf_ilen;
  
  
  sc = (struct ppp_softc *) tp->t_sc;
  
  if (sc == NULL || tp != (struct tty *) sc->sc_devp)
    return;
  
  sc->sc_stats.ppp_ibytes++;
  
  c &= 0xff;
  
  if (c & 0x80)
    sc->sc_flags |= SC_RCV_B7_1;
  else
    sc->sc_flags |= SC_RCV_B7_0;
  if (paritytab[c >> 5] & (1 << (c & 0x1F)))
    sc->sc_flags |= SC_RCV_ODDP;
  else
    sc->sc_flags |= SC_RCV_EVNP;
  
  if (sc->sc_flags & SC_LOG_RAWIN)
    ppplogchar(sc, c);
  
  /*
   * If we receive the flag, do the processing to make
   * sure that the characters we have been receiving make
   * up a valid packet.  If they do, we put that packet out
   * to the system and then we wake up the top half routines
   * that may be waiting on that packet (i.e. we are the bottom
   * half routine and we wake up pppread which is the top half
   * routine).  We return when done.
   *
   * If sc_fill_inprogress is NULL, then we have been flushing
   * characters. Check to see netbufs and
   * if so, allocate and start on a new
   * packet.  Remember, inside this if() statment is the only
   * place we can reset the SC_FLUSH flag.
   *
   */
  
  if (c == PPP_FLAG)
    {
      
      /*
       * We have just received a complete packet.
       * Try to make sure we have enough input
       * buffers so that our future allocations
       * will work.
       *
       * Thanks to Miron Cupermans excellent suggestion/example
       * the interrupt allocation problem has been solved.
       */

      if (sc->sc_empty_recv_netbufs.len < EMPTY_NETBUF_MIN)
	kern_serv_callout(&instance, top_off_recv_netbuf_queue, (void*) sc);
	  

      /*
       * We flush characters if we don't have
       * a receive buffer.  Here, we check to see
       * if we can now allocate one.  If not,
       * return with flush flag still set. Else
       * reset and start all over.
       */
      
      if ((sc->sc_fill_inprogress == NULL) ||
	  (sc->sc_flags & SC_FLUSH))
	{
	  if_ierrors_set(sc->sc_if, if_ierrors(sc->sc_if) + 1);
	  sc->sc_stats.ppp_ierrors++;

	  /*
	   * Try to reallocate
	   */
	  if(sc->sc_fill_inprogress != NULL)
	    NB_FREE(sc->sc_fill_inprogress);
	  
	  sc->sc_fill_inprogress = netbuf_dequeue(&sc->sc_empty_recv_netbufs, 0);

	  if (sc->sc_fill_inprogress == NULL)
	    {
	      sc->sc_flags |= SC_FLUSH;     /* probably redundant */
	      if(sc->sc_flags & SC_LOG_FLUSH)
		printf("No input netbuf...Flushing next incoming packet\n");
	      return;
	    }
	  else
	    {
	      if (sc->sc_flags & SC_LOG_FLUSH)
		printf("Disabling Flush... continuing... \n");
	      sc->sc_nbuf_ilen = 0;
	      sc->sc_flags &= ~(SC_FLUSH | SC_ESCAPED);
	      return;
	    }
	}

      /*
       * If we get here, we probably have a netbuf full
       * of data.  Maybe bad data, but date none the less
       * If we get a flush FF we just return.
       */

      nbuf = sc->sc_fill_inprogress;
      nbuf_ilen = sc->sc_nbuf_ilen;

      if (nbuf_ilen == 0)
	return;


      /*
       * Grow netbuf back to proper size.
       */

      nb_grow_top (nbuf, nbuf_ilen);
      
      if (sc->sc_rawin_count > 0) 
	ppplogchar(sc, -1);
      
#ifdef ADD_ERRORS
      /*
       * This section will add (very poor pseudo) random packet
       * errors if defined.
       */

      if (error_limit == -1)
	{
	  error_limit = 150;
	  printf("Introducing random error after %d more packets\n",
		 error_limit);
	}	

      if (error_pkt_count >=  error_limit)
	{
	  sc->sc_fcs = PPP_GOODFCS - 1;

	  error_pkt_count = 0;
	  error_limit = (int) ((error_limit + 23) & 0x000000ff) + 23;
	  printf("Introducing random error after %d more packets\n",
		 error_limit);
	}
      else
	++error_pkt_count;
      

#endif /* ADD_ERRORS */      


      /*
       * If SC_ESCAPED is set, then we've seen the packet
       * abort sequence "}~".
       */
      
      if ((sc->sc_flags & SC_ESCAPED) ||
	  (nbuf_ilen > 0 && sc->sc_fcs != PPP_GOODFCS))
	{
	  if (nbuf_ilen > 0 && sc->sc_fcs != PPP_GOODFCS)
	    {
	      if (sc->sc_flags & SC_LOG_FLUSH)
		printf("ppp%d: bad frame checksum calculated... dropping packet\n", if_unit(sc->sc_if));
	      if_ierrors_set(sc->sc_if, if_ierrors(sc->sc_if) + 1);
	      sc->sc_stats.ppp_ierrors++;
	    }
	  
	  if (nbuf)
	    NB_FREE(nbuf);
	  
	  sc->sc_fill_inprogress = netbuf_dequeue(&sc->sc_empty_recv_netbufs, 0);
	  sc->sc_nbuf_ilen = 0;
	  sc->sc_flags &= ~(SC_FLUSH | SC_ESCAPED);
	  
	  if (sc->sc_fill_inprogress == NULL)
	    {
	      sc->sc_flags |= SC_FLUSH;
	      if(sc->sc_flags & SC_LOG_FLUSH)
		printf("No input netbuf...Flushing next incoming packet\n");
	    }
	  
	  return;
	}
      
      /*
       * Now reset for next incoming packet. We check the return
       * value later
       */
      sc->sc_nbuf_ilen = 0;
      sc->sc_fill_inprogress = netbuf_dequeue(&sc->sc_empty_recv_netbufs, 0);
      sc->sc_flags &= ~(SC_FLUSH | SC_ESCAPED);
      
      if (sc->sc_fill_inprogress == NULL)
	{
	  if(sc->sc_flags & SC_LOG_FLUSH)
	    printf("No input netbufs...Flushing all incoming packets\n");
	  sc->sc_flags |= SC_FLUSH;
	}
      

      if (nbuf_ilen < PPP_HDRLEN + PPP_FCSLEN)
	{
	  if (sc->sc_flags & SC_DEBUG)
	    printf("ppp%d: too short (%d)\n", if_unit(sc->sc_if), nbuf_ilen);
	  if_ierrors_set(sc->sc_if, if_ierrors(sc->sc_if) + 1);
	  sc->sc_stats.ppp_ierrors++;

	  if(nbuf)
	    NB_FREE(nbuf);

	  return;
	}
      
      /*
       * Remove FCS trailer.
       */
      
      nbuf_ilen -= 2;
      ASSERT((nb_size(nbuf) - nbuf_ilen) >= 2);
      nb_shrink_bot(nbuf, nb_size(nbuf) - (nbuf_ilen));
      
      
      if (sc->sc_flags & SC_LOG_INPKT)
	{
	  printf("ppp%d: got a good packet of %d bytes\n", if_unit(sc->sc_if), nbuf_ilen);
	  ppp_dump_netbuf(nbuf, nbuf_ilen);
	}
      

      if (ppppktin(sc, nbuf, nbuf_ilen))
	{
	  /*
	   * We get here if packet was successfully
	   * placed on sc_inq input, then we wakeup any user processes
	   * that are sleeping on a read.  pppread will take off the
	   * placeholder.
	   */
	  
	  /* Put a placeholder byte in canq for ttselect()/ttnread(). */
	  putc(0, &tp->t_canq);
	  
	  ttwakeup(tp);
	}
      
      return;
      
    } /* c == PPP_FLAG */
  
  /*
   * Only a PPP_FLAG can reset SC_FLUSH
   * So if we get here then the character is not
   * PPP_FLAG and we just drop character.
   */
  
  if (sc->sc_flags & SC_FLUSH)
    {
      if (sc->sc_flags & SC_LOG_FLUSH)
	ppplogchar(sc, c);
      return;
    }
  
  if (c < 0x20 && (sc->sc_rasyncmap & (1 << c)))
    return;
  
  if (sc->sc_flags & SC_ESCAPED)
    {
      sc->sc_flags &= ~SC_ESCAPED;
      c ^= PPP_TRANS;                 /* previous was escape so fix this one */
    }
  else if (c == PPP_ESCAPE)
    {
      sc->sc_flags |= SC_ESCAPED;     /* set up so next time we fix character */
      return;
    }
  
  /*
   * By this time, we have a valid unescaped
   * character to deal with. However, it does
   * not complete a PPP packet so we need to put it
   * in our netbuf.  If we get here, then we
   * also have a valid netbuf.
   *
   * Initialize buffer on first octet received.
   * First octet could be address or protocol (when compressing
   * address/control).
   * Second octet is control.
   * Third octet is first or second (when compressing protocol)
   * octet of protocol.
   * Fourth octet is second octet of protocol.
   */
  
  ASSERT(sc->sc_fill_inprogress != NULL);
  
  if (sc->sc_nbuf_ilen == 0)
    {
      /* reset the first input netbuf */
      
      nbuf = sc->sc_fill_inprogress;

      sc->sc_fcs = PPP_INITFCS;
      if (c != PPP_ALLSTATIONS)
	{
	  if (sc->sc_flags & SC_REJ_COMP_AC)
	    {
	      if (sc->sc_flags & SC_DEBUG)
		printf("ppp%d: garbage received: 0x%x (need PPP_ALLSTATIONS 0xFF)\n",
		       if_unit(sc->sc_if), c);
	      goto flush;
	    }

	  ASSERT(nb_size(sc->sc_fill_inprogress) >= 2);

	  cur_pos = (u_char *) nb_map(sc->sc_fill_inprogress);
	  *cur_pos++ = PPP_ALLSTATIONS;
	  *cur_pos++ = PPP_UI;
	  sc->sc_nbuf_ilen += 2;
	  nb_shrink_top(nbuf, 2);
	}
    }
  
  if (sc->sc_nbuf_ilen == 1 && c != PPP_UI)
    {
      if (sc->sc_flags & SC_DEBUG)
	printf("ppp%d: missing UI (0x3), got 0x%x\n",
	       if_unit(sc->sc_if), c);
      goto flush;
    }
  
  if (sc->sc_nbuf_ilen == 2 && (c & 1) == 1) 
    {
      
      /* a compressed protocol */
      
      ASSERT(nb_size(sc->sc_fill_inprogress) >= 1);
      cur_pos = (u_char *) nb_map(sc->sc_fill_inprogress);
      *cur_pos++ = 0;
      nb_shrink_top(sc->sc_fill_inprogress, 1); 
      sc->sc_nbuf_ilen++;
    }
  
  
  if (sc->sc_nbuf_ilen == 3 && (c & 1) == 0)
    {
      
      if (sc->sc_flags & SC_DEBUG)
	{
	  cur_pos = (u_char *) nb_map(sc->sc_fill_inprogress) - 1;
	  printf("ppp%d: bad protocol %x\n", if_unit(sc->sc_if),
		 (*cur_pos << 8) + c);
	}
      goto flush;
    }
  
  /* packet beyond configured mru? */
  
  if (sc->sc_nbuf_ilen >= sc->sc_mru + PPP_HDRLEN + PPP_FCSLEN)
    {
      if (sc->sc_flags & SC_DEBUG)
	printf("ppp%d: packet too big\n", if_unit(sc->sc_if));
      goto flush;
    }
  
  
  nbuf = sc->sc_fill_inprogress;

  /*
   * This should be impossible, but lets be paranoid.
   */
  if (nbuf == NULL)
    {
      printf("Impossibility...Netbuf was NULL at %s %d\n", __FILE__, __LINE__);
      printf("Flushing... send bug report to perkins@cps.msu.edu!\n");
      goto flush;
    }
  
  /*
   * Make sure we have room for next character.
   */

  if (nb_size(nbuf) < 1)
    {
      if (sc->sc_flags & SC_DEBUG)
	printf("ppp%d: netbuf too small\n", if_unit(sc->sc_if));
      goto flush;
    }
  
  /*
   * Finally, add the character in! 
   *
   */
  
  cur_pos = (u_char *) nb_map(nbuf);
  if (cur_pos == NULL)
    {
      printf("cur_pos was NULL at %s %d\n", __FILE__,__LINE__);
      printf("Send bug report to perkins@cps.msu.edu\n");
      goto flush;
    }
  else
    {
      *cur_pos++ = c;
      nb_shrink_top(nbuf, 1);
      ++sc->sc_nbuf_ilen;
      sc->sc_fcs = PPP_FCS(sc->sc_fcs, c);
    }
  
  return;
  
 flush:
  if (sc->sc_fill_inprogress != NULL)
    {
      NB_FREE(sc->sc_fill_inprogress);
      sc->sc_fill_inprogress = NULL;
      sc->sc_nbuf_ilen = 0;
    }
  if (!(sc->sc_flags & SC_FLUSH))
    {

      sc->sc_flags |= SC_FLUSH;
      if (sc->sc_flags & SC_LOG_FLUSH)
	{
	  printf("FLUSHING....\n");
	  ppplogchar(sc, c);
	}
    }
}




/*
 * Start output on async tty interface.  Get another datagram
 * to send from the interface queue and start sending it.
 *
 * Must be called at interrupt priorority.
 * 
 */
void
pppstart(tp)
    register struct tty *tp;
{
    register struct ppp_softc *sc = (struct ppp_softc *) tp->t_sc;
    register int len;
    register u_char *start, *stop, *cp;
    int n, ndone, done;
    netbuf_t nbuf;

    if ((tp->t_state & TS_CARR_ON) == 0 && (tp->t_flags & CLOCAL) == 0)
      {
	if (sc->sc_flags & SC_DEBUG)
	  {
	    printf("Packet to be sent but modem appears\n");
	    printf("to be down at %s %d\n", __FILE__, __LINE__);
	  }
	return;
      }

    if (sc == NULL || tp != (struct tty *) sc->sc_devp)
      {
	printf("Didn't think we'd get here %s %d\n", __FILE__, __LINE__);
	(*tp->t_oproc)(tp);
	return;
      }

    while(1)
      {
	/*
	 * If there is more in the output queue, just send it now.
	 * We are being called in lieu of ttstart and must do what
	 * it would.
	 *
	 * Seems that there is an incompatibility with the Mux
	 * driver and the character outqueue.  Aparently the
	 * driver doesn't keep proper track of the number
	 * of characters actually waiting to be sent (at least
	 * where we can get ahold of the value).  The PPP_HIWAT
	 * test will be of dubious value on systems using the
	 * Mux driver.
	 *
	 * Should probably put a conditional compilation flag
	 * here, but I don't want to have to worry about this
	 * at the moment.
	 *
	 */

/*	if (CCOUNT(&tp->t_outq) != 0 && tp->t_oproc != NULL) */

	if (tp->t_oproc != NULL)
	  {
	    (*tp->t_oproc)(tp);  

	    if (CCOUNT(&tp->t_outq) > PPP_HIWAT)
		return;
	  }

	/*
	 * See if we have an existing packet partly sent.
	 * If not, get a new packet and start sending it.
	 * We take packets on the priority queue ahead of those
	 * on the normal queue.
	 */
	nbuf = sc->sc_send_inprogress;

	if (nbuf == NULL)
	  {
	    /*
	     * Get another packet to be sent.  Gets it off the correct
             * queue.   Packets were placed on the queue by pppoutput.
	     */
	    nbuf = ppp_dequeue(sc);

	    if (nbuf == NULL)
		return;

	    /*
	     * The extra PPP_FLAG will start up a new packet, and thus
	     * will flush any accumulated garbage.  We do this whenever
	     * the line may have been idle for some time.
	     */
	    if (CCOUNT(&tp->t_outq) == 0)
	      {
		++sc->sc_stats.ppp_obytes;
		(void) putc(PPP_FLAG, &tp->t_outq);
	      }

	    /* Calculate the FCS for this netbuf's worth. */

	    sc->sc_outfcs = pppfcs(PPP_INITFCS, nb_map(nbuf), nb_size(nbuf)); 
	  }


	/*
	 * By this time we have a dequeued netbuf.
	 * We start picking off characters and stuffing them
	 * into the serial driver.  We quit when the serial
	 * driver gets full.  
	 *
	 */

	    len = nb_size(nbuf);
	    start = nb_map(nbuf);
	    stop = start + len;


           /*
	    * While there are more chararcters in the nbuf
	    */
	 
	    while (len > 0)
	      {
		/*
		 * Find out how many bytes in the string we can
		 * handle without doing something special.
		 */
		for (cp = start; cp < stop; cp++)
		    if (ESCAPE_P(*cp))
			break;

		n = cp - start;

		/*
		 * Here, n is the number of characters we
		 * can put out without special processing.
		 * ndone is the number actually output.
		 */
		if (n)
		  {
		    ndone = n - b_to_q(start, n, &tp->t_outq);
		    len -= ndone;
		    
		    start += ndone;
		    
		    sc->sc_stats.ppp_obytes += ndone;

		    /*
		     * ndone was the number that fit into
		     * the serial driver.  n was the number
		     * of characters we could handle before
		     * doing something special.  If we
		     * break here, then we had more valid
		     * (unspecial) characters than would fit.
		     */
		  
		    if (ndone < n)
			break;	/* packet doesn't fit */
		  }
		/*
		 * If there are characters left in the netbuf,
		 * the first one must be special..
		 * Put it out in a different form.  We break
		 * out of the while loop the first time a character
		 * doesn't fit into the serial driver.
		 */

		if (len)
		  {
		    if (putc(PPP_ESCAPE, &tp->t_outq))
			break;
		    if (putc(*start ^ PPP_TRANS, &tp->t_outq))
		      {
			(void) unputc(&tp->t_outq);
			break;
		      }

		    sc->sc_stats.ppp_obytes += 2;
		    start++;

		    len--;

		  }

	      } /* while (len > 0) */


	    /*
	     * We can get to here in one of two ways.
	     * Either the serial driver couldn't handle anymore characters.
	     * In that case, we put the unfinished buffer in a holding
	     * place.
	     *
	     * The other case is that we finished the netbuf so we try to put out
	     * the FCS.
	     *
	     * If we didn't empty this netbuf, remember where we're up to.
	     * If we emptied it, try to add the FCS and closing
	     * flag, and if we can't, leave sc_send_inprogress pointing to us, but with
	     * length == 0, to remind us to output the FCS and flag later.
	     */

	    done = len == 0;

	    if (done)
	      {
		u_char *p, *q;
		int c;
		u_char endseq[8];

		/*
		 * We may have to escape the bytes in the FCS.
		 */
		p = endseq;
		c = ~sc->sc_outfcs & 0xFF;
		if (ESCAPE_P(c))
		  {
		    *p++ = PPP_ESCAPE;
		    *p++ = c ^ PPP_TRANS;
		}
		else
		    *p++ = c;
		c = (~sc->sc_outfcs >> 8) & 0xFF;
		if (ESCAPE_P(c))
		  {
		    *p++ = PPP_ESCAPE;
		    *p++ = c ^ PPP_TRANS;
		}
		else
		    *p++ = c;
		*p++ = PPP_FLAG;

		/*
		 * Try to output the FCS and flag.  If the bytes
		 * don't all fit, back out.
		 */
		for (q = endseq; q < p; ++q)
		    if (putc(*q, &tp->t_outq))
		      {
			done = 0;
			for (; q > endseq; --q)
			    unputc(&tp->t_outq);
			break;
		      }
	      }  /* done */

	    /*
	     * Here, save the netbuf for next time around.
	     */
	    
	    if (!done)
	      {
		nb_shrink_top(nbuf, nb_size(nbuf) - len);
		sc->sc_send_inprogress = nbuf; 
		
		if (tp->t_oproc != NULL)
		    (*tp->t_oproc)(tp);

		return;		/* can't do any more at the moment */
	      }

	/* Finished a packet */

	sc->sc_send_inprogress = NULL;

	sc->sc_stats.ppp_obytes++; /* account for closing flag */

	if_opackets_set(sc->sc_if, if_opackets(sc->sc_if) + 1);
	sc->sc_stats.ppp_opackets++;

	NB_FREE(nbuf);  /* get rid */

    } /* while (1) */

}  /* pppstart() */



/*
 * This gets called from pppoutput when a new packet is
 * put on a queue.  Basically just calls pppstart with a different
 * argumant.
 */

int
pppasyncstart(sc)
    register struct ppp_softc *sc;
{
    register struct tty *tp = (struct tty *) sc->sc_devp;
    int s;

    s = splimp();
    pppstart(tp);
    splx(s);
    return 0;
}


/*
 * Calculate a new FCS given the current FCS and the new data.
 */
u_short
pppfcs(fcs, cp, len)
    register u_short fcs;
    register u_char *cp;
    register int len;
{
    while (len--)
	fcs = PPP_FCS(fcs, *cp++);
    return (fcs);
}



void ppp_dump_netbuf(nbuf, pktlen)
     netbuf_t nbuf;
    int pktlen;
{
    char buf[3*MAX_DUMP_BYTES+4];
    char *bp = buf;
    int len;
    u_char *rptr;
    static char digits[] = "0123456789abcdef";

    len = MIN(nb_size(nbuf), pktlen);
    rptr = (u_char *) nb_map(nbuf);

    

	while (len--)
	  {
	    if (bp > buf + sizeof(buf) - 4)
	      break;
	    *bp++ = digits[*rptr >> 4]; /* convert byte to ascii hex */
	    *bp++ = digits[*rptr++ & 0xf];
	  }

    if (nb_size(nbuf) > pktlen)
	*bp++ = '>';
    *bp = 0;

    printf("%s\n", buf);
}

void pppdumpb(b, l)
    u_char *b;
    int l;
{
    char buf[3*MAX_DUMP_BYTES+4];
    char *bp = buf;
    static char digits[] = "0123456789abcdef";

    while (l--) {
	if (bp >= buf + sizeof(buf) - 3) {
	    *bp++ = '>';
	    break;
	}
	*bp++ = digits[*b >> 4]; /* convert byte to ascii hex */
	*bp++ = digits[*b++ & 0xf];
	*bp++ = ' ';
    }

    *bp = 0;
    printf("%s\n", buf);
}


void ppplogchar(sc, c)
    struct ppp_softc *sc;
    int c;
{
    if (c >= 0)
	sc->sc_rawin[sc->sc_rawin_count++] = c;
    if (sc->sc_rawin_count >= sizeof(sc->sc_rawin)
	|| c < 0 && sc->sc_rawin_count > 0) {
	printf("ppp%d input: ", if_unit(sc->sc_if));
	pppdumpb(sc->sc_rawin, sc->sc_rawin_count);
	sc->sc_rawin_count = 0;
    }
}



